# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest

from datetime import datetime
from hl7apy.base_datatypes import DT, TM, DTM, ST, FT, ID, IS, TX, GTS, NM, SI
from hl7apy.factories import datatype_factory
from hl7apy.exceptions import InvalidDateFormat, InvalidDateOffset, MaxLengthReached, InvalidHighlightRange, InvalidDataType
from hl7apy.validation import VALIDATION_LEVEL
from decimal import Decimal
import numbers

class TestDatatypeFactory(unittest.TestCase):

    #Datatype factory test case

    def test_datatype_creation(self):
        st = datatype_factory('ST', 'string')
        self.assertEqual(st.classname, 'ST')

    def test_datatype_creation_strict(self):
        tx = datatype_factory('TX', 'text', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertEqual(tx.classname, 'TX')

    def test_invalid_datatype_creation(self):
        self.assertRaises(InvalidDataType, datatype_factory, 'XXX', 'text')

    def test_datatype_not_allowed_value_creation_strict(self):
        self.assertRaises(ValueError, datatype_factory, 'TM', '999999', validation_level=VALIDATION_LEVEL.STRICT)

    def test_datatype_not_allowed_value_creation(self):
        datatype_factory('TM', '9999')

    def test_create_timestamp_with_offset_by_factory(self):
        tm = datatype_factory('TM', '120000+0000')
        self.assertEqual(tm.classname, 'TM')
        self.assertEqual(tm.offset, '+0000')

    def test_create_year_month_date_by_factory(self):
        dt = datatype_factory('DT', '201201')
        self.assertEqual(dt.classname, 'DT')
        self.assertEqual(dt.to_er7(), '201201')

    def test_create_year_month_day_date_by_factory(self):
        dt = datatype_factory('DT', '20120101')
        self.assertEqual(dt.classname, 'DT')
        self.assertEqual(dt.to_er7(), '20120101')

    def test_create_hour_time_by_factory(self):
        tm = datatype_factory('TM', '14')
        self.assertEqual(tm.classname, 'TM')
        self.assertEqual(tm.to_er7(), '14')

    def test_create_nm_by_factory(self):
        nm = datatype_factory('NM', 100000)
        self.assertEqual(nm.classname, 'NM')
        self.assertEqual(nm.to_er7(), '100000')

    def test_create_invalid_nm_by_factory_strict(self):
        self.assertRaises(ValueError, datatype_factory, 'NM', 'aaaaa', validation_level=VALIDATION_LEVEL.STRICT)



class TestDT(unittest.TestCase):

    #Test DT datatype

    def test_DT(self):
        date = datetime.strptime('2013', '%Y')
        date1 = datetime.strptime('201307', '%Y%m')
        date2 = datetime.strptime('20130715', '%Y%m%d')
        dt = DT(date, format='%Y')
        dt1 = DT(date1, format='%Y%m')
        dt2 = DT(date2, format= '%Y%m%d')
        self.assertEqual(dt.classname, 'DT')
        self.assertEqual(dt.to_er7(), datetime.strftime(date, dt.format))
        self.assertEqual(dt1.to_er7(), datetime.strftime(date1, dt1.format))
        self.assertEqual(dt2.to_er7(), datetime.strftime(date2, dt2.format))


    def test_DT_wrong_format(self):
        self.assertRaises(InvalidDateFormat, DT,datetime.strptime('13', '%y'), format='%y')
        self.assertRaises(InvalidDateFormat, DT,datetime.strptime('12', '%m'), format='%m')
        self.assertRaises(InvalidDateFormat, DT,datetime.strptime('1202', '%m%d'), format='%m%d')
        self.assertRaises(InvalidDateFormat, DT,datetime.strptime('02', '%d'), format='%d')

    def test_DT_default_format(self):
        date = datetime.strptime('20130715', '%Y%m%d')
        dt = DT(date)
        self.assertEqual(dt.to_er7(), datetime.strftime(date, dt.format))

    def test_DT_wrong_default_format_strict(self):
        date = datetime.strptime('20', '%H')
        self.assertRaises(InvalidDateFormat, DT, date, format='%H', )


class TestTM(unittest.TestCase):

    #Test TM datatype

    def test_TM(self):
        time = datetime.strptime('01', '%H')
        time1 = datetime.strptime('0101', '%H%M')
        time2 = datetime.strptime('010111', '%H%M%S')
        time3 = datetime.strptime('010111.111', '%H%M%S.%f')
        tm = TM(time, format='%H')
        tm1 = TM(time1, format='%H%M')
        tm2 = TM(time2, format='%H%M%S')
        tm3 = TM(time3, format='%H%M%S.%f')
        self.assertEqual(tm.classname, 'TM')
        self.assertEqual(tm.to_er7(), datetime.strftime(time, tm.format))
        self.assertEqual(tm1.to_er7(), datetime.strftime(time1, tm1.format))
        self.assertEqual(tm2.to_er7(), datetime.strftime(time2, tm2.format))
        self.assertEqual(tm3.to_er7(), datetime.strftime(time3, tm3.format))

    def test_TM_wrong_format(self):
        tm = datetime.strptime('01', '%I')
        self.assertRaises(InvalidDateFormat, TM, tm, format='%I')
        self.assertRaises(InvalidDateFormat, TM,datetime.strptime('12', '%M'), format='%M')
        self.assertRaises(InvalidDateFormat, TM,datetime.strptime('1212', '%M%S'), format='%M%S')


    def test_TM_default_format(self):
        time = datetime.strptime('010111.111', '%H%M%S.%f')
        tm = TM(time)
        self.assertEqual(tm.to_er7(), datetime.strftime(time, tm.format))

    def test_TM_offset(self):
        time = datetime.strptime('0101', '%H%M')
        time2 = datetime.strptime('010111.111', '%H%M%S.%f')
        tm = TM(time, format='%H%M', offset='+0100')
        tm2 = TM(time2, format='%H%M%S.%f', offset='-0300')
        TM(time, offset="+0000")
        TM(time, offset="-0000")
        TM(time, offset="+1200")
        TM(time, offset="-1200")
        self.assertEqual(tm.offset, '+0100')
        self.assertEqual(tm2.offset, '-0300')
        self.assertEqual(tm.to_er7(), '0101+0100') #check if a space is needed between time and offset, or not
        self.assertEqual(tm2.to_er7(), '010111.111000-0300')

    def test_TM_invalid_offset(self):
        self.assertRaises(InvalidDateOffset, TM, datetime.strptime('0101', '%H%M'), offset = '+00:00')
        self.assertRaises(InvalidDateOffset, TM, datetime.strptime('0101', '%H%M'), offset = '&0100')
        self.assertRaises(InvalidDateOffset, TM, datetime.strptime('0101', '%H%M'), offset = '+100')
        self.assertRaises(InvalidDateOffset, TM, datetime.strptime('0101', '%H%M'), offset = '+1300')
        self.assertRaises(InvalidDateOffset, TM, datetime.strptime('0101', '%H%M'), offset = '-1300')


class TestDTM(unittest.TestCase):

    def test_DTM(self):
        dtime  = datetime.strptime('2013', '%Y')
        dtime1 = datetime.strptime('201307', '%Y%m')
        dtime2 = datetime.strptime('20130715', '%Y%m%d')
        dtime3 = datetime.strptime('01', '%H')
        dtime4 = datetime.strptime('0101', '%H%M')
        dtime5 = datetime.strptime('010111', '%H%M%S')
        dtime6 = datetime.strptime('010111.111', '%H%M%S.%f')
        dtime7 = datetime.strptime('20130715010111.111', '%Y%m%d%H%M%S.%f')

        dtm  = DTM(dtime, format='%Y')
        dtm1 = DTM(dtime1, format='%Y%m')
        dtm2 = DTM(dtime2, format= '%Y%m%d')
        dtm3 = TM(dtime3, format= '%H')
        dtm4 = TM(dtime4, format=  '%H%M')
        dtm5 = TM(dtime5, format= '%H%M%S')
        dtm6 = TM(dtime6, format= '%H%M%S.%f')
        dtm7 = DTM(dtime7, format= '%Y%m%d%H%M%S.%f')

        self.assertEqual(dtm.classname, 'DTM')
        self.assertEqual(dtm.to_er7(), datetime.strftime(dtime, dtm.format))
        self.assertEqual(dtm1.to_er7(), datetime.strftime(dtime1, dtm1.format))
        self.assertEqual(dtm2.to_er7(), datetime.strftime(dtime2, dtm2.format))
        self.assertEqual(dtm3.to_er7(), datetime.strftime(dtime3, dtm3.format))
        self.assertEqual(dtm4.to_er7(), datetime.strftime(dtime4, dtm4.format))
        self.assertEqual(dtm5.to_er7(), datetime.strftime(dtime5, dtm5.format))
        self.assertEqual(dtm6.to_er7(), datetime.strftime(dtime6, dtm6.format))
        self.assertEqual(dtm7.to_er7(), datetime.strftime(dtime7, dtm7.format))
        self.assertEqual(dtm7.to_er7(), datetime.strftime(dtime7, dtm7.format))

    def test_DTM_default_format(self):
        dtime = datetime.strptime('20130715 010111.111', '%Y%m%d %H%M%S.%f')
        dtm = DTM(dtime)
        self.assertEqual(dtm.to_er7(), datetime.strftime(dtime, dtm.format))

    def test_DTM_wrong_format(self):
        dtime = datetime.strptime('0715010111.111', '%m%d%H%M%S.%f')
        self.assertRaises(InvalidDateFormat, DTM, dtime, format='%m%d%H%M%S.%f')

    def test_DTM_offset(self):
        dtime =  dtime = datetime.strptime('20130715 010111.111000', '%Y%m%d %H%M%S.%f')
        dtm = DTM(dtime, offset = '+0100')
        self.assertEqual(dtm.to_er7(),'20130715010111.111000+0100' )

    def test_DTM_invalid_offset(self):
        self.assertRaises(InvalidDateOffset, DTM, datetime.strptime('2013', '%Y'), offset = '+00:00')
        self.assertRaises(InvalidDateOffset, DTM, datetime.strptime('2013', '%Y'), offset = '&0100')
        self.assertRaises(InvalidDateOffset, DTM, datetime.strptime('2013', '%Y'), offset = '+100')


class TestST(unittest.TestCase):

    #Test ST datatype

    def test_ST(self):
        st = ST('Specimen')
        self.assertEqual(st.classname, 'ST')
        self.assertEqual(st.to_er7(), 'Specimen')

    def test_ST_maxlength_strict(self):
        str = 'a' * (ST(' ').max_length +1)
        self.assertRaises(MaxLengthReached, ST, str, validation_level=VALIDATION_LEVEL.STRICT)

    def test_ST_maxlength(self):
        str = 'a' * (ST(' ').max_length +100)
        st = ST(str) #no exception is raised here

    def test_ST_field_escaping(self):
        st = ST('|field|')
        self.assertEqual(st.to_er7(), '\\F\\field\\F\\')

    def test_ST_component_escape(self):
        st = ST('^component^')
        self.assertEqual(st.to_er7(), '\\S\\component\\S\\')

    def test_ST_subcomponent_escape(self):
        st = ST('&subcomponent&')
        self.assertEqual(st.to_er7(), '\\T\\subcomponent\\T\\')

    def test_ST_repetition_escape(self):
        st = ST('~repetition~')
        self.assertEqual(st.to_er7(), '\\R\\repetition\\R\\')

    def test_ST_highlights(self):
        st = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (18,24)))
        self.assertEqual(st.to_er7(), '\\H\\HIGHLIGHTED\\N\\TEXTIMP\\H\\ORTANT\\N\\')

    def test_ST_invalid_highlight_range(self):
        st = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (4,24)))
        self.assertRaises(InvalidHighlightRange, st.to_er7)
        st = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((4,24), (0,11)))
        self.assertRaises(InvalidHighlightRange, st.to_er7)
        st = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((5,11), (0,11)))
        self.assertRaises(InvalidHighlightRange, st.to_er7)
        st = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (0,4)))
        self.assertRaises(InvalidHighlightRange, st.to_er7)


class TestFT(unittest.TestCase):

    #Test FT datatype

    def test_FT(self):
        text = 'This is a FT datatype text'
        ft = FT(text)
        self.assertEqual(ft.classname, ('FT'))
        self.assertEqual(ft.to_er7(), text)

    def test_FT_maxlength_strict(self):
        ft = 'a'*(FT(' ').max_length +1)
        self.assertRaises(MaxLengthReached, FT, ft, validation_level=VALIDATION_LEVEL.STRICT)

    def test_FT_maxlength(self):
        ft_str = 'a'*(FT(' ').max_length +1)
        ft = FT(ft_str) #no exception is raised here

    def test_FT_field_escaping(self):
        ft = FT('|field|')
        self.assertEqual(ft.to_er7(), '\\F\\field\\F\\')

    def test_FT_component_escape(self):
        ft = FT('^component^')
        self.assertEqual(ft.to_er7(), '\\S\\component\\S\\')

    def test_FT_subcomponent_escape(self):
        ft = FT('&subcomponent&')
        self.assertEqual(ft.to_er7(), '\\T\\subcomponent\\T\\')

    def test_FT_repetition_escape(self):
        ft = FT('~repetition~')
        self.assertEqual(ft.to_er7(), '\\R\\repetition\\R\\')

    def test_FT_highlights(self):
        ft = FT('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (18,24)))
        self.assertEqual(ft.to_er7(), '\\H\\HIGHLIGHTED\\N\\TEXTIMP\\H\\ORTANT\\N\\')

    def test_FT_invalid_highlight_range(self):
        ft = FT('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (4,24)))
        self.assertRaises(InvalidHighlightRange, ft.to_er7)
        ft = FT('HIGHLIGHTEDTEXTIMPORTANT', highlights=((4,24), (0,11)))
        self.assertRaises(InvalidHighlightRange, ft.to_er7)
        ft = FT('HIGHLIGHTEDTEXTIMPORTANT', highlights=((5,11), (0,11)))
        self.assertRaises(InvalidHighlightRange, ft.to_er7)
        ft = FT('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (0,4)))
        self.assertRaises(InvalidHighlightRange, ft.to_er7)


class TestID(unittest.TestCase):

    #Test ID datatype

    def test_ID(self):
        id_str = 'This is a ID'
        id = ID(id_str)
        self.assertEqual(id.classname, ('ID'))
        self.assertEqual(id.to_er7(), id_str)

    #TODO: According to further implementations, add some tests to check strict validations

    #Test IS datatype

    def test_IS(self):
        is_str = 'This is a IS'
        is_obj = IS(is_str)
        self.assertEqual(is_obj.classname, ('IS'))
        self.assertEqual(is_obj.to_er7(), is_str)

    def test_IS_maxlength_strict(self):
        is_str = 'a'*(IS(' ').max_length +1)
        self.assertRaises(MaxLengthReached, IS, is_str, validation_level=VALIDATION_LEVEL.STRICT)

    #TODO: the same as IS


class TestTX(unittest.TestCase):

    #Test TX datatype

    def test_TX(self):
        text = 'This is a TX datatype text'
        tx = TX(text)
        self.assertEqual(tx.classname, ('TX'))
        self.assertEqual(tx.to_er7(), text)

    def test_TX_maxlength_strict(self):
        tx = 'a'*(TX(' ').max_length +1)
        self.assertRaises(MaxLengthReached, TX, tx, validation_level=VALIDATION_LEVEL.STRICT)

    def test_TX_maxlength(self):
        tx_str = 'a'*(TX(' ').max_length +1)
        ft = TX(tx_str) #no exception is raised here

    def test_TX_field_escaping(self):
        tx = TX('|field|')
        self.assertEqual(tx.to_er7(), '\\F\\field\\F\\')

    def test_TX_component_escape(self):
        tx = TX('^component^')
        self.assertEqual(tx.to_er7(), '\\S\\component\\S\\')

    def test_TX_subcomponent_escape(self):
        tx = TX('&subcomponent&')
        self.assertEqual(tx.to_er7(), '\\T\\subcomponent\\T\\')

    def test_TX_repetition_escape(self):
        tx = TX('~repetition~')
        self.assertEqual(tx.to_er7(), '\\R\\repetition\\R\\')

    def test_TX_highlights(self):
        tx = TX('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (18,24)))
        self.assertEqual(tx.to_er7(), '\\H\\HIGHLIGHTED\\N\\TEXTIMP\\H\\ORTANT\\N\\')

    def test_TX_invalid_highlight_range(self):
        tx = TX('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (4,24)))
        self.assertRaises(InvalidHighlightRange, tx.to_er7)
        tx = TX('HIGHLIGHTEDTEXTIMPORTANT', highlights=((4,24), (0,11)))
        self.assertRaises(InvalidHighlightRange, tx.to_er7)
        tx = TX('HIGHLIGHTEDTEXTIMPORTANT', highlights=((5,11), (0,11)))
        self.assertRaises(InvalidHighlightRange, tx.to_er7)
        tx = TX('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (0,4)))
        self.assertRaises(InvalidHighlightRange, tx.to_er7)


class TestGTS(unittest.TestCase):

    #Test GTS Datatype

    def test_GTS(self):
        text = 'This is a GTS datatype text'
        gts = GTS(text)
        self.assertEqual(gts.classname, 'GTS')
        self.assertEqual(gts.to_er7(), text)

    def test_GTS_maxlength_strict(self):
        gts = 'a'*(GTS(' ').max_length +1)
        self.assertRaises(MaxLengthReached, GTS, gts, validation_level=VALIDATION_LEVEL.STRICT)

    def test_GTS_maxlength(self):
        gts_str = 'a'*(GTS(' ').max_length +1)
        ft = GTS(gts_str) #no exception is raised here

    def test_GTS_field_escaping(self):
        gts = GTS('|field|')
        self.assertEqual(gts.to_er7(), '\\F\\field\\F\\')

    def test_GTS_component_escape(self):
        gts = GTS('^component^')
        self.assertEqual(gts.to_er7(), '\\S\\component\\S\\')

    def test_GTS_subcomponent_escape(self):
        gts = GTS('&subcomponent&')
        self.assertEqual(gts.to_er7(), '\\T\\subcomponent\\T\\')

    def test_GTS_repetition_escape(self):
        gts = GTS('~repetition~')
        self.assertEqual(gts.to_er7(), '\\R\\repetition\\R\\')

    def test_GTS_highlights(self):
        gts = GTS('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (18,24)))
        self.assertEqual(gts.to_er7(), '\\H\\HIGHLIGHTED\\N\\TEXTIMP\\H\\ORTANT\\N\\')

    def test_GTS_invalid_highlight_range(self):
        gts = GTS('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (4,24)))
        self.assertRaises(InvalidHighlightRange, gts.to_er7)
        gts = GTS('HIGHLIGHTEDTEXTIMPORTANT', highlights=((4,24), (0,11)))
        self.assertRaises(InvalidHighlightRange, gts.to_er7)
        gts = GTS('HIGHLIGHTEDTEXTIMPORTANT', highlights=((5,11), (0,11)))
        self.assertRaises(InvalidHighlightRange, gts.to_er7)
        gts = GTS('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (0,4)))
        self.assertRaises(InvalidHighlightRange, gts.to_er7)


class TestNM(unittest.TestCase):

    #Test NM datatype

    def test_NM(self):
        num = NM(Decimal(1234))
        self.assertEqual(num.classname, 'NM')
        self.assertEqual(num.to_er7(),'1234')
        num = 1234
        nm = NM(num)
        num = float(1234.54)
        nm = NM(num)

    def test_NM_maxlength_strict(self):
        num = Decimal(1000000000000000000000000000000000000000000000000000000000000000000000000000)
        self.assertRaises(MaxLengthReached, NM, num, validation_level=VALIDATION_LEVEL.STRICT)

    def test_NM_maxlength(self):
        num = Decimal(1000000000000000000000000000000000000000000000000000000000000000000000000000)
        nm = NM(num)
        self.assertEqual(nm.to_er7(), '1000000000000000000000000000000000000000000000000000000000000000000000000000')


    def test_invalid_NM(self):
        self.assertRaises(ValueError, NM, 'aaaaa')


class TestSI(unittest.TestCase):

    #Test SI datatype


    def test_SI(self):
        si = SI(1234)
        self.assertEqual(si.classname, 'SI')
        self.assertEqual(si.to_er7(), '1234')

    def test_SI_maxlength_strict(self):
        si = 10000
        self.assertRaises(MaxLengthReached, SI, si, validation_level=VALIDATION_LEVEL.STRICT)

    def test_SI_maxlength(self):
        SI(10000) # no exception is raised here

    def test_invalid_SI(self):
        self.assertRaises(ValueError, SI, 'aaaaaa')



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import uuid

from hl7apy.core import Message, Segment
from hl7apy.parser import parse_message

class LB(object):
    """
    Label broker implementation
    """
    @staticmethod
    def send(patient_id):
        # create a QBP_Q11 query message for the given patient_id
        m = Message("QBP_Q11")
        m.MSH.sending_application = "LB module" # same as m.msh.msh_3
        m.MSH.receiving_application = "LIP module" # same as m.msh.msh_5
        m.MSH.MSH_9 = "QBP^SLI^QBP_Q11"
        m.MSH.MSH_10 = uuid.uuid4().hex
        m.QPD.QPD_1 = "SLI^Specimen Labeling Instructions^IHE_LABTF"
        m.QPD.query_tag = uuid.uuid4().hex
        m.QPD.QPD_3 = patient_id
        m.RCP = "RCP|I||R"
        return m.to_er7()

    @staticmethod
    def receive(message):
        # print to stdout the data received
        print "Received by LB"
        try:
            # parse the incoming HL7 message
            m = parse_message(message, find_groups=False)
        except:
            print 'parsing failed', repr(message)
        else:
            print "Message type:", m.MSH.message_type.to_er7()
            print "Message content:", repr(m.to_er7())
            surname, name = m.PID.PID_5.family_name.to_er7(), m.PID.PID_5.given_name.to_er7()
            print "Patient data:", surname, name

class LIP(object):
    """
    Label information provider implementation
    """

    @staticmethod
    def nak(message):
        """
        Build a NAK response for the incoming message

        :param message: incoming message
        :return: a NAK message
        """
        response = Message("ACK")
        response.MSH.MSH_9 = "ACK"
        response.MSA.MSA_1 = "AE"
        response.MSA.MSA_2 = message.MSH.MSH_10
        response.MSA.MSA_3 = "Message type not supported"
        return response

    @staticmethod
    def reply(message):
        """
        Parse the incoming message and return the ER7 encoded response

        :param message: incoming message
        :return: response encoded to ER7 - it can be a NAK or RSP_K11 message
        """
        print "Received by LIP", repr(message)

        try:
            # parse the incoming message
            m = parse_message(message, find_groups=False)
        except:
            print 'parsing failed', repr(message)
            response = LIP.nak()
        else:
            print "Message type:", m.MSH.message_type.to_er7()
            print "Message content:", repr(m.to_er7())
            if m.MSH.MSH_9.MSH_9_3.to_er7() == 'QBP_Q11':
                # create a new RSP_K11 message
                response = Message("RSP_K11")
                response.MSH.MSH_9 = "RSP^K11^RSP_K11"
                # add MSA segment
                response.MSA = "MSA|AA"
                response.MSA.MSA_2 = m.MSH.MSH_10
                # create a QAK segment
                qak = Segment("QAK")
                qak.qak_2 = m.QPD.QPD_2
                qak.qak_3 = "OK"
                qak.qak_4 = "Q22^Specimen Labeling Instructions^IHE_LABTF"
                qak.qak_5 = "1"
                # add the QAK segment to the RSP_K11 message
                response.add(qak)
                # copy the QPD segment from the incoming message
                response.QPD = m.QPD
                # create a PID segment
                response.PID.PID_1 = '1'
                response.PID.PID_5.PID_5_1 = 'PATIENT_SURNAME'
                response.PID.PID_5.PID_5_2 = 'PATIENT_NAME'
                response.PID.PID_6 = "19800101"
                response.PID.PID_7 = "F"
                # create a SPM segment
                spm = Segment("SPM")
                # create an OBR segment
                obr = Segment("OBR")
                spm.SPM_1 = '1'
                spm.SPM_2 = "12345"
                obr.OBR_4 = "ORDER^DESCRIPTION"
                # add spm and obr to the RSP_K11 response
                response.add(spm)
                response.add(obr)
            else:
                response = LIP.nak(m)
        return response.to_er7() # encode to ER7



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
HL7apy - parsing functions
"""

import re

from hl7apy import get_default_encoding_chars, get_default_version, check_version, check_encoding_chars
from hl7apy.consts import N_SEPS
from hl7apy.core import is_base_datatype, Message, Group, Segment, Field, Component, SubComponent, ElementFinder
from hl7apy.exceptions import InvalidName, ParserError, InvalidEncodingChars
from hl7apy.validation import Validator

def parse_message(message, validation_level=None, find_groups=True, reference=None):
    """
    Parse the given ER7-encoded message and return an instance of :class:`hl7apy.core.Message`.

    :type message: ``basestring``
    :param message: the ER7-encoded message to be parsed
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :type find_groups: ``bool``
    :param find_groups: if ``True``, automatically assign the segments found to the appropriate :class:`hl7apy.core.Group` instances. If ``False``, the segments found are assigned as children of the :class:`hl7apy.core.Message` instance
    :return: an instance of :class:`hl7apy.core.Message`

    >>> message = "MSH|^~\&|GHH_ADT||||20080115153000||OML^O33^OML_O33|0123456789|P|2.5||||AL\\rPID|1||566-554-3423^^^GHH^MR||EVERYMAN^ADAM^A|||M|||2222 HOME STREET^^ANN ARBOR^MI^^USA||555-555-2004|||M\\r"
    >>> m = parse_message(message)
    >>> print m
    <Message OML_O33>
    >>> print m.msh.sending_application.to_er7()
    GHH_ADT
    >>> print m.children
    [<Segment MSH>, <Group OML_O33_PATIENT>]
    """
    message = message.lstrip()
    encoding_chars, message_structure, version = get_message_info(message)

    try:
        m = Message(name=message_structure, version=version, validation_level=validation_level, encoding_chars=encoding_chars)
    except InvalidName:
        m = Message(version=version, validation_level=validation_level, encoding_chars=encoding_chars)
    children = parse_segments(message, m.version, encoding_chars, validation_level)
    if m.name is not None and find_groups:
        m.children = []
        create_groups(m, children, validation_level)
    else:
        m.children = children
    if Validator.is_strict(validation_level):
        m.validate()
    return m

def parse_segments(text, version=None, encoding_chars=None, validation_level=None):
    """
    Parse the given ER7-encoded segments and return a list of :class:`hl7apy.core.Segment` instances.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the segments to be parsed
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :return: a list of :class:`hl7apy.core.Segment` instances

    >>> segments = "EVN||20080115153000||||20080114003000\\rPID|1||566-554-3423^^^GHH^MR||EVERYMAN^ADAM^A|||M|||2222 HOME STREET^^ANN ARBOR^MI^^USA||555-555-2004|||M\\r"
    >>> print parse_segments(segments)
    [<Segment EVN>, <Segment PID>]
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    segment_sep = encoding_chars['SEGMENT']
    return [parse_segment(s.strip(), version, encoding_chars, validation_level) for s in text.split(segment_sep) if len(s) > 0]

def parse_segment(text, version=None, encoding_chars=None, validation_level=None, reference=None):
    """
    Parse the given ER7-encoded segment and return an instance of :class:`hl7apy.core.Segment`.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the segment to be parsed
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :type reference: ``dict``
    :param reference: a dictionary containing the element structure returned by :func:`hl7apy.load_reference` or :func:`hl7apy.find_reference`
    :return: an instance of :class:`hl7apy.core.Segment`

    >>> segment = "EVN||20080115153000||||20080114003000"
    >>> s =  parse_segment(segment)
    >>> print s
    <Segment EVN>
    >>> print s.to_er7()
    EVN||20080115153000||||20080114003000
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    segment_name = text[:3]
    text = text[4:] if segment_name != 'MSH' else text[3:]
    segment = Segment(segment_name, version=version, validation_level=validation_level, reference=reference)
    segment.children = parse_fields(text, segment_name, version, encoding_chars, validation_level, segment.allow_infinite_children)
    return segment

def parse_fields(text, name_prefix=None, version=None, encoding_chars=None, validation_level=None, force_varies=False):
    """
    Parse the given ER7-encoded fields and return a list of :class:`hl7apy.core.Field`.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the fields to be parsed
    :type name_prefix: ``basestring``
    :param name_prefix: the field prefix (e.g. MSH)
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :return: a list of :class:`hl7apy.core.Field` instances

    >>> fields = "1|NUCLEAR^NELDA^W|SPO|2222 HOME STREET^^ANN ARBOR^MI^^USA"
    >>> nk1_fields = parse_fields(fields, name_prefix="NK1")
    >>> print nk1_fields
    [<Field NK1_1 (SET_ID_NK1) of type SI>, <Field NK1_2 (NAME) of type XPN>, <Field NK1_3 (RELATIONSHIP) of type CE>, <Field NK1_4 (ADDRESS) of type XAD>]
    >>> s = Segment("NK1")
    >>> s.children = nk1_fields
    >>> print s.to_er7()
    NK1|1|NUCLEAR^NELDA^W|SPO|2222 HOME STREET^^ANN ARBOR^MI^^USA
    >>> unknown_fields = parse_fields(fields)
    >>> s.children = unknown_fields
    >>> print s.to_er7()
    NK1|1|NUCLEAR^NELDA^W|SPO|2222 HOME STREET^^ANN ARBOR^MI^^USA
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    text = text.strip("\r")
    field_sep = encoding_chars['FIELD']
    repetition_sep = encoding_chars['REPETITION']
    splitted_fields = text.split(field_sep)
    fields = []
    for index, field in enumerate(splitted_fields):
        name = "{0}_{1}".format(name_prefix, index+1) if name_prefix is not None else None
        if field.strip() or name is None:
            if name == 'MSH_2':
                fields.append(parse_field(field, name, version, encoding_chars, validation_level))
            else:
                for rep in field.split(repetition_sep):
                    fields.append(parse_field(rep, name, version, encoding_chars, validation_level, force_varies=force_varies))
        elif name == "MSH_1":
            fields.append(parse_field(field_sep, name, version, encoding_chars, validation_level))
    return fields

def parse_field(text, name=None, version=None, encoding_chars=None, validation_level=None, reference=None, force_varies=False):
    """
    Parse the given ER7-encoded field and return an instance of :class:`hl7apy.core.Field`.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the fields to be parsed
    :type name: ``basestring``
    :param name: the field name (e.g. MSH_7)
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :type reference: ``dict``
    :param reference: a dictionary containing the element structure returned by :func:`hl7apy.load_reference` or :func:`hl7apy.find_reference`
    :return: an instance of :class:`hl7apy.core.Field`

    >>> field = "NUCLEAR^NELDA^W"
    >>> nk1_2 = parse_field(field, name="NK1_2")
    >>> print nk1_2
    <Field NK1_2 (NAME) of type XPN>
    >>> print nk1_2.to_er7()
    NUCLEAR^NELDA^W
    >>> unknown = parse_field(field)
    >>> print unknown
    <Field of type None>
    >>> print unknown.to_er7()
    NUCLEAR^NELDA^W
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    if force_varies:
        reference = ('leaf', 'varies', None, None)
    try:
        field = Field(name, version=version, validation_level=validation_level, reference=reference)
    except InvalidName:
        field = Field(version=version, validation_level=validation_level, reference=reference)
    if name in ('MSH_1', 'MSH_2'):
        s = SubComponent(datatype='ST', value=text)
        c = Component(datatype='ST')
        c.add(s)
        field.add(c)
    else:
        children = parse_components(text, field.datatype, version, encoding_chars, validation_level)
        if Validator.is_quiet(validation_level) and is_base_datatype(field.datatype, version) and \
                len(children) > 1:
            field.datatype = None
        field.children = children
    return field

def parse_components(text, field_datatype='ST', version=None, encoding_chars=None, validation_level=None):
    """
    Parse the given ER7-encoded components and return a list of :class:`hl7apy.core.Component` instances.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the components to be parsed
    :type field_datatype: ``basestring``
    :param field_datatype: the datatype of the components (e.g. ST)
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :return: a list of :class:`hl7apy.core.Component` instances

    >>> components = "NUCLEAR^NELDA^W^^TEST"
    >>> xpn = parse_components(components, field_datatype="XPN")
    >>> print xpn
    [<Component XPN_1 (FAMILY_NAME) of type FN>, <Component XPN_2 (GIVEN_NAME) of type ST>, <Component XPN_3 (SECOND_AND_FURTHER_GIVEN_NAMES_OR_INITIALS_THEREOF) of type ST>, <Component XPN_5 (PREFIX_E_G_DR_) of type ST>]
    >>> print parse_components(components)
    [<Component ST (None) of type ST>, <Component ST (None) of type ST>, <Component ST (None) of type ST>, <Component ST (None) of type ST>, <Component ST (None) of type ST>]
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    component_sep = encoding_chars['COMPONENT']
    components = []
    for index, component in enumerate(text.split(component_sep)):
        if is_base_datatype(field_datatype, version):
            component_datatype = field_datatype
            component_name = None
        elif field_datatype is None or field_datatype == 'varies':
            component_datatype = None
            component_name = 'VARIES_{0}'.format(index+1)
        else:
            component_name = "{0}_{1}".format(field_datatype, index+1)
            component_datatype = None
        if component.strip() or component_name is None or component_name.startswith("VARIES_"):
            components.append(parse_component(component, component_name, component_datatype,
                                              version, encoding_chars, validation_level))
    return components

def parse_component(text, name=None, datatype='ST', version=None, encoding_chars=None, validation_level=None, reference=None):
    """
    Parse the given ER7-encoded component and return an instance of :class:`hl7apy.core.Component`.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the components to be parsed
    :type name: ``basestring``
    :param name: the component's name (e.g. XPN_2)
    :type datatype: ``basestring``
    :param datatype: the datatype of the component (e.g. ST)
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :type reference: ``dict``
    :param reference: a dictionary containing the element structure returned by :func:`hl7apy.load_reference` or :func:`hl7apy.find_reference`
    :return: an instance of :class:`hl7apy.core.Component`

    >>> component = "GATEWAY&1.3.6.1.4.1.21367.2011.2.5.17"
    >>> cx_4 = parse_component(component, name="CX_4")
    >>> print cx_4
    <Component CX_4 (ASSIGNING_AUTHORITY) of type None>
    >>> print cx_4.to_er7()
    GATEWAY&1.3.6.1.4.1.21367.2011.2.5.17
    >>> print parse_component(component)
    <Component ST (None) of type None>
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    try:
        component = Component(name, datatype, version=version, validation_level=validation_level, reference=reference)
    except InvalidName as e:
        if Validator.is_strict(validation_level):
            raise e
        component = Component(datatype, version=version, validation_level=validation_level, reference=reference)
    children = parse_subcomponents(text, component.datatype, version, encoding_chars, validation_level)
    if Validator.is_quiet(component.validation_level) and is_base_datatype(component.datatype, version) and \
            len(children) > 1:
        component.datatype = None
    component.children = children
    return component

def parse_subcomponents(text, component_datatype='ST', version=None, encoding_chars=None, validation_level=None):
    """
    Parse the given ER7-encoded subcomponents and return a list of :class:`hl7apy.core.SubComponent` instances.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the components to be parsed
    :type component_datatype: ``basestring``
    :param component_datatype: the datatype of the subcomponents (e.g. ST)
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :return: a list of :class:`hl7apy.core.SubComponent` instances

    >>> subcomponents= "ID&TEST&&AHAH"
    >>> cwe = parse_subcomponents(subcomponents, component_datatype="CWE")
    >>> print cwe
    [<SubComponent CWE_1>, <SubComponent CWE_2>, <SubComponent CWE_4>]
    >>> c = Component(datatype='CWE')
    >>> c.children = cwe
    >>> print c.to_er7()
    ID&TEST&&AHAH
    >>> subs = parse_subcomponents(subcomponents)
    >>> print subs
    [<SubComponent ST>, <SubComponent ST>, <SubComponent ST>, <SubComponent ST>]
    >>> c.children = subs
    >>> print c.to_er7()
    ID&TEST&&AHAH
    """
    version = _get_version(version)
    encoding_chars = _get_encoding_chars(encoding_chars)

    subcomp_sep = encoding_chars['SUBCOMPONENT']
    subcomponents = []
    for index, subcomponent in enumerate(text.split(subcomp_sep)):
        if is_base_datatype(component_datatype, version) or component_datatype is None:
            subcomponent_name = None
            subcomponent_datatype = component_datatype if component_datatype is not None else 'ST'
        else:
            subcomponent_name = "{0}_{1}".format(component_datatype, index+1)
            subcomponent_datatype = None
        if subcomponent.strip() or subcomponent_name is None:
            subcomponents.append(parse_subcomponent(subcomponent, subcomponent_name, subcomponent_datatype, version, validation_level))
    return subcomponents

def parse_subcomponent(text, name=None, datatype='ST', version=None, validation_level=None):
    """
    Parse the given ER7-encoded component and return an instance of :class:`hl7apy.core.Component`.

    :type text: ``basestring``
    :param text: the ER7-encoded string containing the subcomponent data
    :type name: ``basestring``
    :param name: the subcomponent's name (e.g. XPN_2)
    :type datatype: ``basestring``
    :param datatype: the datatype of the subcomponent (e.g. ST)
    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :return: an instance of :class:`hl7apy.core.SubComponent`
    """
    return SubComponent(name=name, datatype=datatype, value=text, version=version, validation_level=validation_level)

def get_message_info(content):
    """
    Parse the given ER7-encoded message to find its version, structure and encoding chars

    :param content: the ER7-encoded message
    :return: a tuple containing (encoding_chars, message_structure, version)
    """
    regex = re.compile("^MSH(?P<field_sep>\S){1}")
    m = regex.match(content)
    if m is not None: # if the regular expression matches, it is an HL7 message
        field_sep = m.group('field_sep') # get the field separator (first char after MSH)
        msh = content.split("\r", 1)[0] # get the first segment
        fields = msh.split(field_sep)
        seps = fields[1] # get the remaining encoding chars (MSH.2)
        if len(seps) > len(set(seps)):
            raise InvalidEncodingChars("Found duplicate encoding chars")
        try:
           comp_sep, rep_sep, escape, sub_sep = seps
        except ValueError:
            if len(seps) < N_SEPS:
                raise InvalidEncodingChars('Missing required encoding chars')
            else:
                raise InvalidEncodingChars('Found {0} encoding chars'.format(len(seps)))
        else:
            encoding_chars = {
                'FIELD': field_sep,
                'COMPONENT': comp_sep,
                'SUBCOMPONENT': sub_sep,
                'REPETITION': rep_sep,
                'ESCAPE': escape,
                'SEGMENT': '\r',
                'GROUP': '\r',
            }
    else:
        raise ParserError("Invalid message")

    # look for MSH.9 field (e.g. ADT^A01^ADT_A01) containing the message structure
    try:
        msh_9 = fields[8].strip()
    except IndexError:
        message_structure = None
    else:
        message_type = msh_9.split(encoding_chars['COMPONENT'])
        try:
            message_structure = message_type[2]
        except IndexError:
            try:
                message_structure = "{0}_{1}".format(message_type[0], message_type[1])
            except IndexError:
                message_structure = None

    # look for MSH.12 field containing the message's version
    try:
        msh_12 = fields[11].strip()
    except IndexError:
        version = None
    else:
        version_id = msh_12.split(encoding_chars['COMPONENT'])
        try:
            version = version_id[0]
        except:
            version = None

    return encoding_chars, message_structure, version

def create_groups(message, children, validation_level=None):
    """
    Create the message's groups

    :param message: a :class:`hl7apy.core.Message` instance
    :param children: a list of :class:`hl7apy.core.Segment` instances
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    """
    # get the message structure
    structure = ElementFinder.get_structure(message)
    # create the initial search data structure
    search_data = {'parents': [message], 'indexes': [-1], 'structures': [structure]}
    # for each segment found in the message...
    for c in children:
        found = -1
        for x in xrange(len(search_data['structures'])):
            found = _find_group(c, search_data, validation_level)
            # group not found at the current level, go back to the previous level
            if found == -1:
                _go_back(search_data)
            else:
                break
        # group not found: add the segment to the message instance
        if found == -1:
            message.add(c)

def _go_back(search_data):
    # go back to the previous level during group search
    search_data['parents'].pop()
    search_data['structures'].pop()
    search_data['indexes'].pop()

def _find_group(segment, search_data, validation_level=None):
    """
    Recursively find the group the segment belongs to

    :param segment: a :class:`hl7apy.core.Segment` instance
    :param search_data: a dictionary containing current chain of parent, current indexes and element structures
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    :return: the index of the parent children list where the segment has been found
    """
    search_index = -1
    current_index = search_data['indexes'][-1]
    structure = search_data['structures'][-1]
    try:
        search_index = structure['structure_by_name'].keys().index(segment.name)
    except ValueError:
        # groups of the current parent
        groups = [k for k, v in structure['structure_by_name'].iteritems() if v['cls'] == Group]
        # for any group found, create the group and check if the segment is one of its children
        for g in groups:
            group = Group(g, validation_level=validation_level)
            p_structure = ElementFinder.get_structure(group)
            search_data['structures'].append(p_structure)
            search_data['parents'].append(group)
            search_data['indexes'].append(-1)
            found_index = _find_group(segment, search_data, validation_level)
            # the segment is a child of the current group
            if found_index > -1:
                find_parent = search_data['parents'].index(group) - 1
                group.parent = search_data['parents'][find_parent]
                search_index = structure['structure_by_name'].keys().index(g)
                search_data['indexes'][find_parent] = search_index
                break
            else: # the segment is not a child of the current group, continue the search
                _go_back(search_data)
    else:
        if search_index <= current_index and structure['repetitions'][segment.name][1] == 1:
        # if more than one instance of the segment has been found and only one instance is allowed,
        # go up of one level to create another instance of the group the segment belongs to
            _go_back(search_data)
            return _find_group(segment, search_data, validation_level)
        search_data['indexes'][-1] = search_index
        parent = search_data['parents'][-1]
        parent.add(segment)
    return search_index

def _get_version(version):
    if version is None:
        version = get_default_version()
    check_version(version)
    return version

def _get_encoding_chars(encoding_chars):
    if encoding_chars is None:
        encoding_chars = get_default_encoding_chars()
    check_encoding_chars(encoding_chars)
    return encoding_chars


if __name__ == '__main__':

    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest
from hl7apy.core import Message, Segment, Field, Component, SubComponent
from hl7apy.parser import parse_segment
from hl7apy.base_datatypes import *
from hl7apy.exceptions import InvalidHighlightRange

class ToStringTestCase(unittest.TestCase):
    """
    TestCase class which tests the functionalities of the API converting the
    HL7 elements to string
    """
    def setUp(self):
        self.custom_encoding_chars = {'SEGMENT' : '\r',
                                      'FIELD' :  '!',
                                      'COMPONENT' : '$',
                                      'SUBCOMPONENT' : '@',
                                      'REPETITION' : 'r',
                                      'ESCAPE' : '?'}
        self.msh_values_standard = ['|', '^~\&', 'LIP', 'LIP', 'LB', 'LB', '20111207121030', '', 'RSP^SLI^RSP_K11', '430102569308795840', 'D', '2.5', '', '', '', '', 'IT', '', 'EN', '', '']
        self.msh_standard = 'MSH|^~\\&|LIP|LIP|LB|LB|20111207121030||RSP^SLI^RSP_K11|430102569308795840|D|2.5|||||IT||EN'
        self.msh_values_custom = ['!', '$r?@', 'LIP', 'LIP', 'LB', 'LB', '20111207121030', '', 'RSP^SLI^RSP_K11', '430102569308795840', 'D', '2.5', '', '', '', '', 'IT', '', 'EN', '', '']
        self.msh_custom = 'MSH!$r?@!LIP!LIP!LB!LB!20111207121030!!RSP^SLI^RSP_K11!430102569308795840!D!2.5!!!!!IT!!EN'
        self.msh_highlighted = 'MSH|^~\\&|LIP|LIP|LB|LB|20111207121030|\\H\\HIGHLIGHTED\\N\\TEXT\\H\\IMPORTANT\\N\\|RSP^SLI^RSP_K11|430102569308795840|D|2.5|||||IT||EN'

    def _get_msh(self, values):
        encoding_chars = {'SEGMENT' : '\r',
                          'FIELD' :  values[0],
                          'COMPONENT' : values[1][0],
                          'SUBCOMPONENT' : values[1][1],
                          'REPETITION' : values[1][2],
                          'ESCAPE' : values[1][3]}
        msh = parse_segment('MSH{0}{1}'.format(encoding_chars['FIELD'], encoding_chars['FIELD'].join(values[1:])), encoding_chars=encoding_chars)
        return msh

    def _create_test_message(self, msh_values):
        """
        Create a test message - RSP_K11 - with only the msh segment.
        The msh is filled with the sent in input
        """
        msg = Message('RSP_K11')
        msg.msh = self._get_msh(msh_values)
        return msg

    def test_msg_to_string_standard_encoding_chars(self):
        """
        It tests the to_er7 message functionality using default encoding chars
        """
        msg = self._create_test_message(self.msh_values_standard)
        self.assertEqual(msg.to_er7(), self.msh_standard)

    def test_msg_to_string_custom_encoding_chars(self):
        """
        It tests the to_er7 message functionality using custom encoding chars
        """
        msg = self._create_test_message(self.msh_values_custom)
        msg.to_er7(self.custom_encoding_chars)
        self.assertEqual(msg.to_er7(self.custom_encoding_chars), self.msh_custom)

    def test_msg_to_string_empty(self):
        """
        It tests the to_er7 message for an empty message
        """
        from datetime import datetime

        msg = Message('RSP_K11')
        self.assertEqual(msg.to_er7(), 'MSH|^~\\&|||||%s|||||2.5' %datetime.strftime(datetime.now(), '%Y%m%d%H%M%S')
)

    def test_highlights(self):
        """
        It tests the highlighting functionaly
        """
        msg = self._create_test_message(self.msh_values_standard)
        value = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (15,24)))
        s = SubComponent(datatype='ST', value=value)
        c = Component(datatype='ST')
        c.add(s)
        msg.msh.msh_8.msh_8_1 = c
        self.assertEqual(msg.to_er7(), self.msh_highlighted)

        value = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((15,24), (0,11)))
        s = SubComponent(datatype='ST', value=value)
        c = Component(datatype='ST')
        c.add(s)
        msg.msh.msh_8.msh_8_1 = c
        self.assertEqual(msg.to_er7(), self.msh_highlighted)

    def test_invalid_highlights(self):
        """
        It tests that highlighting functionality raises the
        :exc:`InvalidHighlightRange` exception in case of invalid range
        """
        data = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (4,24)))
        self.assertRaises(InvalidHighlightRange, data.to_er7)
        data = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((4,24), (0,11)))
        self.assertRaises(InvalidHighlightRange, data.to_er7)
        data = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((5,11), (0,11)))
        self.assertRaises(InvalidHighlightRange, data.to_er7)
        data = ST('HIGHLIGHTEDTEXTIMPORTANT', highlights=((0,11), (0,4)))
        self.assertRaises(InvalidHighlightRange, data.to_er7)

    def test_to_string_msh_field(self):
        m = Message('OML_O33')
        msh = m.msh
        self.assertEqual(msh.msh_1.to_er7(), '|')
        self.assertEqual(msh.msh_2.to_er7(), '^~\\&')
        msh_1 = Field('MSH_1')
        msh_2 = Field('MSH_2')
        self.assertRaises(IndexError, msh_1.to_er7)
        self.assertRaises(IndexError, msh_2.to_er7)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
HL7apy - Exceptions
"""

class HL7apyException(Exception):
    """
    Base exception class for hl7apy
    """


class ParserError(HL7apyException):
    """
    Error during parsing

    >>> from hl7apy.parser import parse_message
    >>> m = parse_message('NOTHL7')
    Traceback (most recent call last):
    ...
    ParserError: Invalid message
    """


class UnsupportedVersion(HL7apyException):
    """
    Given version is not supported

    >>> from hl7apy import set_default_version
    >>> set_default_version("2.0")
    Traceback (most recent call last):
    ...
    UnsupportedVersion: The version 2.0 is not supported
    """
    def __init__(self, version):
        self.version = version

    def __str__(self):
        return 'The version {0} is not supported'.format(self.version)


class ChildNotFound(HL7apyException):
    """
    Raised when a child element is not found in the HL7 reference structures for the given version

    >>> from hl7apy.core import Segment, Field
    >>> s = Segment('MSH')
    >>> s.unknown = Field()
    Traceback (most recent call last):
    ...
    ChildNotFound: No child named UNKNOWN
    """
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return 'No child named {0}'.format(self.name)


class ChildNotValid(HL7apyException):
    """
    Raised when you try to assign an unexpected child to an :class:`hl7apy.core.Element`

    >>> from hl7apy.core import Segment, Field
    >>> s = Segment('PID', validation_level=1)
    >>> s.pid_1 = Field('PID_34')
    Traceback (most recent call last):
    ...
    ChildNotValid: <Field PID_34 (LAST_UPDATE_FACILITY) of type HD> is not a valid child for PID_1
    """
    def __init__(self, child, parent):
        self.child = child
        self.parent = parent

    def __str__(self):
        return '{0} is not a valid child for {1}'.format(self.child, self.parent)


class UnknownValidationLevel(HL7apyException):
    """
    Raised when the validation_level specified is not valid

    It should be one of those defined in :class:`hl7apy.consts.VALIDATION_LEVEL`.

    >>> from hl7apy import set_default_validation_level
    >>> set_default_validation_level(3)
    Traceback (most recent call last):
    ...
    UnknownValidationLevel
    """


class OperationNotAllowed(HL7apyException):
    """
    Generic exception raised when something is not allowed

    >>> from hl7apy.core import Segment
    >>> s = Segment()
    Traceback (most recent call last):
    ...
    OperationNotAllowed: Cannot instantiate an unknown Segment
    """


class MaxChildLimitReached(HL7apyException):
    """
    Raised when a child cannot be added to an instance of :class:`hl7apy.core.Element`
    since the :class:`hl7apy.core.Element` has already reached the maximum number
    of children allowed for the given child type (e.g. a :class:`hl7apy.core.Message` should have at most 1 MSH segment)

    >>> from hl7apy.core import Message, Segment
    >>> m = Message("OML_O33", validation_level=1)
    >>> m.add(Segment('MSH'))
    Traceback (most recent call last):
    ...
    MaxChildLimitReached: Cannot add <Segment MSH>: max limit (1) reached for <Message OML_O33>
    """
    def __init__(self, parent, child, limit):
        self.parent = parent
        self.child = child
        self.limit = limit

    def __str__(self):
        return 'Cannot add {0}: max limit ({1}) reached for {2}'.format(self.child, self.limit, self.parent)


class MaxLengthReached(HL7apyException):
    """
    Value length exceeds its datatype :attr:`max_length`.

    >>> from hl7apy.v2_5 import get_base_datatypes
    >>> from hl7apy.consts import VALIDATION_LEVEL
    >>> SI = get_base_datatypes()['SI']
    >>> st = SI(value=11111, validation_level=VALIDATION_LEVEL.STRICT)
    Traceback (most recent call last):
    ...
    MaxLengthReached: The value 11111 exceed the max length: 4
    """
    def __init__(self, value, limit):
        self.value = value
        self.limit = limit

    def __str__(self):
        return 'The value {0} exceed the max length: {1}'.format(self.value, self.limit)


class InvalidName(HL7apyException):
    """
    Raised if the reference for the given class/name has not been found

    >>> from hl7apy.core import Message
    >>> Message('Unknown')
    Traceback (most recent call last):
    ...
    InvalidName: Invalid name for Message: UNKNOWN
    """
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name

    def __str__(self):
        return 'Invalid name for {0}: {1}'.format(self.cls, self.name)


class InvalidDataType(HL7apyException):
    """
    Raised when the currently used HL7 version does not support the given datatype

    >>> from hl7apy.factories import datatype_factory
    >>> datatype_factory('TN', '11 123456', version="2.3") #doctest: +ELLIPSIS
    <hl7apy.base_datatypes.TN object at 0x...>
    >>> datatype_factory('TN', '11 123456', version="2.5")
    Traceback (most recent call last):
    ...
    InvalidDataType: The datatype TN is not available for the given HL7 version
    """
    def __init__(self, datatype):
        self.datatype = datatype

    def __str__(self):
        return "The datatype {0} is not available for the given HL7 version".format(self.datatype)


class InvalidHighlightRange(HL7apyException):
    """
    Raised when the specified highlight range is not valid

    For a description of highlight range see :class:`hl7apy.base_datatypes.TextualDataType`

    >>> from hl7apy.v2_5 import get_base_datatypes
    >>> ST = get_base_datatypes()['ST']
    >>> s = ST(value='some useful information', highlights=((5, 3),))
    >>> s.to_er7()
    Traceback (most recent call last):
    ...
    InvalidHighlightRange: Invalid highlight range: 5 - 3
    """
    def __init__(self, lower_bound, upper_bound):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __str__(self):
        return 'Invalid highlight range: {0} - {1}'.format(self.lower_bound, self.upper_bound)


class InvalidDateFormat(HL7apyException):
    """
    Raised when the output format for a :class:`hl7apy.base_datatypes.DateTimeDataType` is not valid

    >>> from hl7apy.v2_5 import get_base_datatypes
    >>> DTM = get_base_datatypes()['DTM']
    >>> DTM(value='10102013', format="%d%m%Y")
    Traceback (most recent call last):
    ...
    InvalidDateFormat: Invalid date format: %d%m%Y
    """
    def __init__(self, format):
        self.format = format

    def __str__(self):
        return 'Invalid date format: {0}'.format(self.format)


class InvalidDateOffset(HL7apyException):
    """
    Raised when the offset for a :class:`TM` or :class:`hl7apy.base_datatypes.DTM` is not valid

    >>> from hl7apy.v2_5 import get_base_datatypes
    >>> DTM = get_base_datatypes()['DTM']
    >>> DTM(value='20131010', format="%Y%m%d", offset='+1300')
    Traceback (most recent call last):
    ...
    InvalidDateOffset: Invalid date offset: +1300
    """
    def __init__(self, offset):
        self.offset = offset

    def __str__(self):
        return 'Invalid date offset: {0}'.format(self.offset)


class InvalidEncodingChars(HL7apyException):
    """
    Raised when the encoding chars specified is not a correct set of HL7 encoding chars

    >>> from hl7apy.core import Message
    >>> encoding_chars = {'GROUP' : '\\r', 'SEGMENT' : '\\r', 'COMPONENT' : '^', 'SUBCOMPONENT' : '&', 'REPETITION' : '~', 'ESCAPE' : '\\\\'}
    >>> m = Message('ADT_A01', encoding_chars=encoding_chars)
    Traceback (most recent call last):
    ...
    InvalidEncodingChars: Invalid encoding chars
    """
    def __str__(self):
        return 'Invalid encoding chars'


if __name__ == '__main__':

    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from distutils.core import setup

setup(
    name = 'hl7apy',
    version = '1.0-rc1',
    description = 'Python library to parse, create and handle HL7 v2.x messages',
    long_description = """
    HL7apy is a Python library to parse, create and handle `HL7 <http://www.hl7.org>`_ v2 messages.

    HL7apy is developed and maintained by `Healthcare Flows <http://www.crs4.it/healthcare-flows>`_ researchers at `CRS4 <http://www.crs4.it>`_.

    Source code on GitHub: https://github.com/crs4/hl7apy

    Documentation can be found here: http://hl7apy.org
    """,
    author = 'Daniela Ghironi, Vittorio Meloni, Alessandro Sulis, Federico Caboni',
    author_email = '<daniela.ghironi@crs4.it>, <vittorio.meloni@crs4.it>, <alessandro.sulis@crs4.it>, <federico.caboni@me.com>',
    url = 'https://github.com/crs4/hl7apy',
    download_url = 'https://github.com/crs4/hl7apy/tarball/v1.0.0-rc.1',
    license = 'MIT License',
    keywords = ['HL7', 'Health Level 7', 'healthcare', 'python'],
    classifiers = [
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ],
    packages = ['hl7apy', 'hl7apy.v2_2', 'hl7apy.v2_3', 'hl7apy.v2_3_1', 'hl7apy.v2_4', 'hl7apy.v2_5', 'hl7apy.v2_5_1', 'hl7apy.v2_6'],
    test_suite = 'tests',
)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

DATATYPES = {'AD': ('sequence',
        (('AD_1', (0, 1)),
         ('AD_2', (0, 1)),
         ('AD_3', (0, 1)),
         ('AD_4', (0, 1)),
         ('AD_5', (0, 1)),
         ('AD_6', (0, 1)),
         ('AD_7', (0, 1)),
         ('AD_8', (0, 1)))),
 'AD_1': ('leaf', 'ST', 'STREET_ADDRESS', None),
 'AD_2': ('leaf', 'ST', 'OTHER_DESIGNATION', None),
 'AD_3': ('leaf', 'ST', 'CITY', None),
 'AD_4': ('leaf', 'ST', 'STATE_OR_PROVINCE', None),
 'AD_5': ('leaf', 'ST', 'ZIP_OR_POSTAL_CODE', None),
 'AD_6': ('leaf', 'ID', 'COUNTRY', None),
 'AD_7': ('leaf', 'ID', 'ADDRESS_TYPE', None),
 'AD_8': ('leaf', 'ST', 'OTHER_GEOGRAPHIC_DESIGNATION', None),
 'AUI': ('sequence',
         (('AUI_1', (0, 1)), ('AUI_2', (0, 1)), ('AUI_3', (0, 1)))),
 'AUI_1': ('leaf', 'ST', 'AUTHORIZATION_NUMBER', None),
 'AUI_2': ('leaf', 'DT', 'DATE', None),
 'AUI_3': ('leaf', 'ST', 'SOURCE', None),
 'CCD': ('sequence', (('CCD_1', (1, 1)), ('CCD_2', (0, 1)))),
 'CCD_1': ('leaf', 'ID', 'INVOCATION_EVENT', None),
 'CCD_2': ('leaf', 'DTM', 'DATE_TIME', None),
 'CCP': ('sequence',
         (('CCP_1', (0, 1)), ('CCP_2', (0, 1)), ('CCP_3', (0, 1)))),
 'CCP_1': ('leaf',
           'NM',
           'CHANNEL_CALIBRATION_SENSITIVITY_CORRECTION_FACTOR',
           None),
 'CCP_2': ('leaf', 'NM', 'CHANNEL_CALIBRATION_BASELINE', None),
 'CCP_3': ('leaf', 'NM', 'CHANNEL_CALIBRATION_TIME_SKEW', None),
 'CD': ('sequence',
        (('CD_1', (0, 1)),
         ('CD_2', (0, 1)),
         ('CD_3', (0, 1)),
         ('CD_4', (0, 1)),
         ('CD_5', (0, 1)),
         ('CD_6', (0, 1)))),
 'CD_1': ('leaf', 'WVI', 'CHANNEL_IDENTIFIER', None),
 'CD_2': ('leaf', 'WVS', 'WAVEFORM_SOURCE', None),
 'CD_3': ('leaf', 'CSU', 'CHANNEL_SENSITIVITY_AND_UNITS', None),
 'CD_4': ('leaf', 'CCP', 'CHANNEL_CALIBRATION_PARAMETERS', None),
 'CD_5': ('leaf', 'NM', 'CHANNEL_SAMPLING_FREQUENCY', None),
 'CD_6': ('leaf', 'NR', 'MINIMUM_AND_MAXIMUM_DATA_VALUES', None),
 'CF': ('sequence',
        (('CF_1', (0, 1)),
         ('CF_2', (0, 1)),
         ('CF_3', (0, 1)),
         ('CF_4', (0, 1)),
         ('CF_5', (0, 1)),
         ('CF_6', (0, 1)))),
 'CF_1': ('leaf', 'ST', 'IDENTIFIER', None),
 'CF_2': ('leaf', 'FT', 'FORMATTED_TEXT', None),
 'CF_3': ('leaf', 'ID', 'NAME_OF_CODING_SYSTEM', None),
 'CF_4': ('leaf', 'ST', 'ALTERNATE_IDENTIFIER', None),
 'CF_5': ('leaf', 'FT', 'ALTERNATE_FORMATTED_TEXT', None),
 'CF_6': ('leaf', 'ID', 'NAME_OF_ALTERNATE_CODING_SYSTEM', None),
 'CNE': ('sequence',
         (('CNE_1', (1, 1)),
          ('CNE_2', (0, 1)),
          ('CNE_3', (0, 1)),
          ('CNE_4', (0, 1)),
          ('CNE_5', (0, 1)),
          ('CNE_6', (0, 1)),
          ('CNE_7', (0, 1)),
          ('CNE_8', (0, 1)),
          ('CNE_9', (0, 1)))),
 'CNE_1': ('leaf', 'ST', 'IDENTIFIER', None),
 'CNE_2': ('leaf', 'ST', 'TEXT', None),
 'CNE_3': ('leaf', 'ID', 'NAME_OF_CODING_SYSTEM', None),
 'CNE_4': ('leaf', 'ST', 'ALTERNATE_IDENTIFIER', None),
 'CNE_5': ('leaf', 'ST', 'ALTERNATE_TEXT', None),
 'CNE_6': ('leaf', 'ID', 'NAME_OF_ALTERNATE_CODING_SYSTEM', None),
 'CNE_7': ('leaf', 'ST', 'CODING_SYSTEM_VERSION_ID', None),
 'CNE_8': ('leaf', 'ST', 'ALTERNATE_CODING_SYSTEM_VERSION_ID', None),
 'CNE_9': ('leaf', 'ST', 'ORIGINAL_TEXT', None),
 'CNN': ('sequence',
         (('CNN_1', (0, 1)),
          ('CNN_2', (0, 1)),
          ('CNN_3', (0, 1)),
          ('CNN_4', (0, 1)),
          ('CNN_5', (0, 1)),
          ('CNN_6', (0, 1)),
          ('CNN_7', (0, 1)),
          ('CNN_8', (0, 1)),
          ('CNN_9', (0, 1)),
          ('CNN_10', (0, 1)),
          ('CNN_11', (0, 1)))),
 'CNN_1': ('leaf', 'ST', 'ID_NUMBER', None),
 'CNN_10': ('leaf', 'ST', 'ASSIGNING_AUTHORITY_UNIVERSAL_ID', None),
 'CNN_11': ('leaf', 'ID', 'ASSIGNING_AUTHORITY_UNIVERSAL_ID_TYPE', None),
 'CNN_2': ('leaf', 'ST', 'FAMILY_NAME', None),
 'CNN_3': ('leaf', 'ST', 'GIVEN_NAME', None),
 'CNN_4': ('leaf',
           'ST',
           'SECOND_AND_FURTHER_GIVEN_NAMES_OR_INITIALS_THEREOF',
           None),
 'CNN_5': ('leaf', 'ST', 'SUFFIX_E_G_JR_OR_III_', None),
 'CNN_6': ('leaf', 'ST', 'PREFIX_E_G_DR_', None),
 'CNN_7': ('leaf', 'IS', 'DEGREE_E_G_MD', None),
 'CNN_8': ('leaf', 'IS', 'SOURCE_TABLE', None),
 'CNN_9': ('leaf', 'IS', 'ASSIGNING_AUTHORITY_NAMESPACE_ID', None),
 'CP': ('sequence',
        (('CP_1', (1, 1)),
         ('CP_2', (0, 1)),
         ('CP_3', (0, 1)),
         ('CP_4', (0, 1)),
         ('CP_5', (0, 1)),
         ('CP_6', (0, 1)))),
 'CP_1': ('leaf', 'MO', 'PRICE', None),
 'CP_2': ('leaf', 'ID', 'PRICE_TYPE', None),
 'CP_3': ('leaf', 'NM', 'FROM_VALUE', None),
 'CP_4': ('leaf', 'NM', 'TO_VALUE', None),
 'CP_5': ('leaf', 'CWE', 'RANGE_UNITS', None),
 'CP_6': ('leaf', 'ID', 'RANGE_TYPE', None),
 'CQ': ('sequence', (('CQ_1', (0, 1)), ('CQ_2', (0, 1)))),
 'CQ_1': ('leaf', 'NM', 'QUANTITY', None),
 'CQ_2': ('leaf', 'CWE', 'UNITS', None),
 'CSU': ('sequence',
         (('CSU_1', (1, 1)),
          ('CSU_2', (0, 1)),
          ('CSU_3', (0, 1)),
          ('CSU_4', (0, 1)),
          ('CSU_5', (0, 1)),
          ('CSU_6', (0, 1)),
          ('CSU_7', (0, 1)))),
 'CSU_1': ('leaf', 'NM', 'CHANNEL_SENSITIVITY', None),
 'CSU_2': ('leaf', 'ST', 'UNIT_OF_MEASURE_IDENTIFIER', None),
 'CSU_3': ('leaf', 'ST', 'UNIT_OF_MEASURE_DESCRIPTION', None),
 'CSU_4': ('leaf', 'ID', 'UNIT_OF_MEASURE_CODING_SYSTEM', None),
 'CSU_5': ('leaf', 'ST', 'ALTERNATE_UNIT_OF_MEASURE_IDENTIFIER', None),
 'CSU_6': ('leaf', 'ST', 'ALTERNATE_UNIT_OF_MEASURE_DESCRIPTION', None),
 'CSU_7': ('leaf', 'ID', 'ALTERNATE_UNIT_OF_MEASURE_CODING_SYSTEM', None),
 'CWE': ('sequence',
         (('CWE_1', (0, 1)),
          ('CWE_2', (0, 1)),
          ('CWE_3', (0, 1)),
          ('CWE_4', (0, 1)),
          ('CWE_5', (0, 1)),
          ('CWE_6', (0, 1)),
          ('CWE_7', (0, 1)),
          ('CWE_8', (0, 1)),
          ('CWE_9', (0, 1)))),
 'CWE_1': ('leaf', 'ST', 'IDENTIFIER', None),
 'CWE_2': ('leaf', 'ST', 'TEXT', None),
 'CWE_3': ('leaf', 'ID', 'NAME_OF_CODING_SYSTEM', None),
 'CWE_4': ('leaf', 'ST', 'ALTERNATE_IDENTIFIER', None),
 'CWE_5': ('leaf', 'ST', 'ALTERNATE_TEXT', None),
 'CWE_6': ('leaf', 'ID', 'NAME_OF_ALTERNATE_CODING_SYSTEM', None),
 'CWE_7': ('leaf', 'ST', 'CODING_SYSTEM_VERSION_ID', None),
 'CWE_8': ('leaf', 'ST', 'ALTERNATE_CODING_SYSTEM_VERSION_ID', None),
 'CWE_9': ('leaf', 'ST', 'ORIGINAL_TEXT', None),
 'CX': ('sequence',
        (('CX_1', (1, 1)),
         ('CX_2', (0, 1)),
         ('CX_3', (0, 1)),
         ('CX_4', (0, 1)),
         ('CX_5', (0, 1)),
         ('CX_6', (0, 1)),
         ('CX_7', (0, 1)),
         ('CX_8', (0, 1)),
         ('CX_9', (0, 1)),
         ('CX_10', (0, 1)))),
 'CX_1': ('leaf', 'ST', 'ID_NUMBER', None),
 'CX_10': ('leaf', 'CWE', 'ASSIGNING_AGENCY_OR_DEPARTMENT', None),
 'CX_2': ('leaf', 'ST', 'IDENTIFIER_CHECK_DIGIT', None),
 'CX_3': ('leaf', 'ID', 'CHECK_DIGIT_SCHEME', None),
 'CX_4': ('leaf', 'HD', 'ASSIGNING_AUTHORITY', None),
 'CX_5': ('leaf', 'ID', 'IDENTIFIER_TYPE_CODE', None),
 'CX_6': ('leaf', 'HD', 'ASSIGNING_FACILITY', None),
 'CX_7': ('leaf', 'DT', 'EFFECTIVE_DATE', None),
 'CX_8': ('leaf', 'DT', 'EXPIRATION_DATE', None),
 'CX_9': ('leaf', 'CWE', 'ASSIGNING_JURISDICTION', None),
 'DDI': ('sequence',
         (('DDI_1', (0, 1)), ('DDI_2', (1, 1)), ('DDI_3', (0, 1)))),
 'DDI_1': ('leaf', 'NM', 'DELAY_DAYS', None),
 'DDI_2': ('leaf', 'MO', 'MONETARY_AMOUNT', None),
 'DDI_3': ('leaf', 'NM', 'NUMBER_OF_DAYS', None),
 'DIN': ('sequence', (('DIN_1', (1, 1)), ('DIN_2', (1, 1)))),
 'DIN_1': ('leaf', 'DTM', 'DATE', None),
 'DIN_2': ('leaf', 'CWE', 'INSTITUTION_NAME', None),
 'DLD': ('sequence', (('DLD_1', (1, 1)), ('DLD_2', (0, 1)))),
 'DLD_1': ('leaf', 'CWE', 'DISCHARGE_TO_LOCATION', None),
 'DLD_2': ('leaf', 'DTM', 'EFFECTIVE_DATE', None),
 'DLN': ('sequence',
         (('DLN_1', (1, 1)), ('DLN_2', (0, 1)), ('DLN_3', (0, 1)))),
 'DLN_1': ('leaf', 'ST', 'LICENSE_NUMBER', None),
 'DLN_2': ('leaf', 'IS', 'ISSUING_STATE_PROVINCE_COUNTRY', None),
 'DLN_3': ('leaf', 'DT', 'EXPIRATION_DATE', None),
 'DLT': ('sequence',
         (('DLT_1', (0, 1)),
          ('DLT_2', (0, 1)),
          ('DLT_3', (0, 1)),
          ('DLT_4', (0, 1)))),
 'DLT_1': ('leaf', 'NR', 'NORMAL_RANGE', None),
 'DLT_2': ('leaf', 'NM', 'NUMERIC_THRESHOLD', None),
 'DLT_3': ('leaf', 'ID', 'CHANGE_COMPUTATION', None),
 'DLT_4': ('leaf', 'NM', 'DAYS_RETAINED', None),
 'DR': ('sequence', (('DR_1', (0, 1)), ('DR_2', (0, 1)))),
 'DR_1': ('leaf', 'DTM', 'RANGE_START_DATE_TIME', None),
 'DR_2': ('leaf', 'DTM', 'RANGE_END_DATE_TIME', None),
 'DTN': ('sequence', (('DTN_1', (1, 1)), ('DTN_2', (1, 1)))),
 'DTN_1': ('leaf', 'IS', 'DAY_TYPE', None),
 'DTN_2': ('leaf', 'NM', 'NUMBER_OF_DAYS', None),
 'ED': ('sequence',
        (('ED_1', (0, 1)),
         ('ED_2', (1, 1)),
         ('ED_3', (0, 1)),
         ('ED_4', (1, 1)),
         ('ED_5', (1, 1)))),
 'ED_1': ('leaf', 'HD', 'SOURCE_APPLICATION', None),
 'ED_2': ('leaf', 'ID', 'TYPE_OF_DATA', None),
 'ED_3': ('leaf', 'ID', 'DATA_SUBTYPE', None),
 'ED_4': ('leaf', 'ID', 'ENCODING', None),
 'ED_5': ('leaf', 'TX', 'DATA', None),
 'EI': ('sequence',
        (('EI_1', (0, 1)),
         ('EI_2', (0, 1)),
         ('EI_3', (0, 1)),
         ('EI_4', (0, 1)))),
 'EIP': ('sequence', (('EIP_1', (0, 1)), ('EIP_2', (0, 1)))),
 'EIP_1': ('leaf', 'EI', 'PLACER_ASSIGNED_IDENTIFIER', None),
 'EIP_2': ('leaf', 'EI', 'FILLER_ASSIGNED_IDENTIFIER', None),
 'EI_1': ('leaf', 'ST', 'ENTITY_IDENTIFIER', None),
 'EI_2': ('leaf', 'IS', 'NAMESPACE_ID', None),
 'EI_3': ('leaf', 'ST', 'UNIVERSAL_ID', None),
 'EI_4': ('leaf', 'ID', 'UNIVERSAL_ID_TYPE', None),
 'ELD': ('sequence',
         (('ELD_1', (0, 1)),
          ('ELD_2', (0, 1)),
          ('ELD_3', (0, 1)),
          ('ELD_4', (0, 1)))),
 'ELD_1': ('leaf', 'ST', 'SEGMENT_ID', None),
 'ELD_2': ('leaf', 'NM', 'SEGMENT_SEQUENCE', None),
 'ELD_3': ('leaf', 'NM', 'FIELD_POSITION', None),
 'ELD_4': ('leaf', 'CWE', 'CODE_IDENTIFYING_ERROR', None),
 'ERL': ('sequence',
         (('ERL_1', (1, 1)),
          ('ERL_2', (1, 1)),
          ('ERL_3', (0, 1)),
          ('ERL_4', (0, 1)),
          ('ERL_5', (0, 1)),
          ('ERL_6', (0, 1)))),
 'ERL_1': ('leaf', 'ST', 'SEGMENT_ID', None),
 'ERL_2': ('leaf', 'NM', 'SEGMENT_SEQUENCE', None),
 'ERL_3': ('leaf', 'NM', 'FIELD_POSITION', None),
 'ERL_4': ('leaf', 'NM', 'FIELD_REPETITION', None),
 'ERL_5': ('leaf', 'NM', 'COMPONENT_NUMBER', None),
 'ERL_6': ('leaf', 'NM', 'SUB_COMPONENT_NUMBER', None),
 'FC': ('sequence', (('FC_1', (1, 1)), ('FC_2', (0, 1)))),
 'FC_1': ('leaf', 'IS', 'FINANCIAL_CLASS_CODE', None),
 'FC_2': ('leaf', 'DTM', 'EFFECTIVE_DATE', None),
 'FN': ('sequence',
        (('FN_1', (1, 1)),
         ('FN_2', (0, 1)),
         ('FN_3', (0, 1)),
         ('FN_4', (0, 1)),
         ('FN_5', (0, 1)))),
 'FN_1': ('leaf', 'ST', 'SURNAME', None),
 'FN_2': ('leaf', 'ST', 'OWN_SURNAME_PREFIX', None),
 'FN_3': ('leaf', 'ST', 'OWN_SURNAME', None),
 'FN_4': ('leaf', 'ST', 'SURNAME_PREFIX_FROM_PARTNER_SPOUSE', None),
 'FN_5': ('leaf', 'ST', 'SURNAME_FROM_PARTNER_SPOUSE', None),
 'HD': ('sequence', (('HD_1', (0, 1)), ('HD_2', (0, 1)), ('HD_3', (0, 1)))),
 'HD_1': ('leaf', 'IS', 'NAMESPACE_ID', None),
 'HD_2': ('leaf', 'ST', 'UNIVERSAL_ID', None),
 'HD_3': ('leaf', 'ID', 'UNIVERSAL_ID_TYPE', None),
 'ICD': ('sequence',
         (('ICD_1', (0, 1)), ('ICD_2', (1, 1)), ('ICD_3', (0, 1)))),
 'ICD_1': ('leaf', 'IS', 'CERTIFICATION_PATIENT_TYPE', None),
 'ICD_2': ('leaf', 'ID', 'CERTIFICATION_REQUIRED', None),
 'ICD_3': ('leaf', 'DTM', 'DATE_TIME_CERTIFICATION_REQUIRED', None),
 'JCC': ('sequence',
         (('JCC_1', (0, 1)), ('JCC_2', (0, 1)), ('JCC_3', (0, 1)))),
 'JCC_1': ('leaf', 'IS', 'JOB_CODE', None),
 'JCC_2': ('leaf', 'IS', 'JOB_CLASS', None),
 'JCC_3': ('leaf', 'TX', 'JOB_DESCRIPTION_TEXT', None),
 'LA1': ('sequence',
         (('LA1_1', (0, 1)),
          ('LA1_2', (0, 1)),
          ('LA1_3', (0, 1)),
          ('LA1_4', (0, 1)),
          ('LA1_5', (0, 1)),
          ('LA1_6', (0, 1)),
          ('LA1_7', (0, 1)),
          ('LA1_8', (0, 1)),
          ('LA1_9', (0, 1)))),
 'LA1_1': ('leaf', 'IS', 'POINT_OF_CARE', None),
 'LA1_2': ('leaf', 'IS', 'ROOM', None),
 'LA1_3': ('leaf', 'IS', 'BED', None),
 'LA1_4': ('leaf', 'HD', 'FACILITY', None),
 'LA1_5': ('leaf', 'IS', 'LOCATION_STATUS', None),
 'LA1_6': ('leaf', 'IS', 'PATIENT_LOCATION_TYPE', None),
 'LA1_7': ('leaf', 'IS', 'BUILDING', None),
 'LA1_8': ('leaf', 'IS', 'FLOOR', None),
 'LA1_9': ('leaf', 'AD', 'ADDRESS', None),
 'LA2': ('sequence',
         (('LA2_1', (0, 1)),
          ('LA2_2', (0, 1)),
          ('LA2_3', (0, 1)),
          ('LA2_4', (0, 1)),
          ('LA2_5', (0, 1)),
          ('LA2_6', (0, 1)),
          ('LA2_7', (0, 1)),
          ('LA2_8', (0, 1)),
          ('LA2_9', (0, 1)),
          ('LA2_10', (0, 1)),
          ('LA2_11', (0, 1)),
          ('LA2_12', (0, 1)),
          ('LA2_13', (0, 1)),
          ('LA2_14', (0, 1)),
          ('LA2_15', (0, 1)),
          ('LA2_16', (0, 1)))),
 'LA2_1': ('leaf', 'IS', 'POINT_OF_CARE', None),
 'LA2_10': ('leaf', 'ST', 'OTHER_DESIGNATION', None),
 'LA2_11': ('leaf', 'ST', 'CITY', None),
 'LA2_12': ('leaf', 'ST', 'STATE_OR_PROVINCE', None),
 'LA2_13': ('leaf', 'ST', 'ZIP_OR_POSTAL_CODE', None),
 'LA2_14': ('leaf', 'ID', 'COUNTRY', None),
 'LA2_15': ('leaf', 'ID', 'ADDRESS_TYPE', None),
 'LA2_16': ('leaf', 'ST', 'OTHER_GEOGRAPHIC_DESIGNATION', None),
 'LA2_2': ('leaf', 'IS', 'ROOM', None),
 'LA2_3': ('leaf', 'IS', 'BED', None),
 'LA2_4': ('leaf', 'HD', 'FACILITY', None),
 'LA2_5': ('leaf', 'IS', 'LOCATION_STATUS', None),
 'LA2_6': ('leaf', 'IS', 'PATIENT_LOCATION_TYPE', None),
 'LA2_7': ('leaf', 'IS', 'BUILDING', None),
 'LA2_8': ('leaf', 'IS', 'FLOOR', None),
 'LA2_9': ('leaf', 'ST', 'STREET_ADDRESS', None),
 'MA': ('sequence',
        (('MA_1', (0, 1)),
         ('MA_2', (0, 1)),
         ('MA_3', (0, 1)),
         ('MA_4', (0, 1)))),
 'MA_1': ('leaf', 'NM', 'SAMPLE_Y_FROM_CHANNEL_1', None),
 'MA_2': ('leaf', 'NM', 'SAMPLE_Y_FROM_CHANNEL_2', None),
 'MA_3': ('leaf', 'NM', 'SAMPLE_Y_FROM_CHANNEL_3', None),
 'MA_4': ('leaf', 'NM', 'SAMPLE_Y_FROM_CHANNEL_4', None),
 'MO': ('sequence', (('MO_1', (0, 1)), ('MO_2', (0, 1)))),
 'MOC': ('sequence', (('MOC_1', (0, 1)), ('MOC_2', (0, 1)))),
 'MOC_1': ('leaf', 'MO', 'MONETARY_AMOUNT', None),
 'MOC_2': ('leaf', 'CWE', 'CHARGE_CODE', None),
 'MOP': ('sequence',
         (('MOP_1', (1, 1)), ('MOP_2', (1, 1)), ('MOP_3', (0, 1)))),
 'MOP_1': ('leaf', 'ID', 'MONEY_OR_PERCENTAGE_INDICATOR', None),
 'MOP_2': ('leaf', 'NM', 'MONEY_OR_PERCENTAGE_QUANTITY', None),
 'MOP_3': ('leaf', 'ID', 'CURRENCY_DENOMINATION', None),
 'MO_1': ('leaf', 'NM', 'QUANTITY', None),
 'MO_2': ('leaf', 'ID', 'DENOMINATION', None),
 'MSG': ('sequence',
         (('MSG_1', (1, 1)), ('MSG_2', (1, 1)), ('MSG_3', (1, 1)))),
 'MSG_1': ('leaf', 'ID', 'MESSAGE_CODE', None),
 'MSG_2': ('leaf', 'ID', 'TRIGGER_EVENT', None),
 'MSG_3': ('leaf', 'ID', 'MESSAGE_STRUCTURE', None),
 'NA': ('sequence',
        (('NA_1', (1, 1)),
         ('NA_2', (0, 1)),
         ('NA_3', (0, 1)),
         ('NA_4', (0, 1)))),
 'NA_1': ('leaf', 'NM', 'VALUE1', None),
 'NA_2': ('leaf', 'NM', 'VALUE2', None),
 'NA_3': ('leaf', 'NM', 'VALUE3', None),
 'NA_4': ('leaf', 'NM', 'VALUE4', None),
 'NDL': ('sequence',
         (('NDL_1', (0, 1)),
          ('NDL_2', (0, 1)),
          ('NDL_3', (0, 1)),
          ('NDL_4', (0, 1)),
          ('NDL_5', (0, 1)),
          ('NDL_6', (0, 1)),
          ('NDL_7', (0, 1)),
          ('NDL_8', (0, 1)),
          ('NDL_9', (0, 1)),
          ('NDL_10', (0, 1)),
          ('NDL_11', (0, 1)))),
 'NDL_1': ('leaf', 'CNN', 'NAME', None),
 'NDL_10': ('leaf', 'IS', 'BUILDING', None),
 'NDL_11': ('leaf', 'IS', 'FLOOR', None),
 'NDL_2': ('leaf', 'DTM', 'START_DATE_TIME', None),
 'NDL_3': ('leaf', 'DTM', 'END_DATE_TIME', None),
 'NDL_4': ('leaf', 'IS', 'POINT_OF_CARE', None),
 'NDL_5': ('leaf', 'IS', 'ROOM', None),
 'NDL_6': ('leaf', 'IS', 'BED', None),
 'NDL_7': ('leaf', 'HD', 'FACILITY', None),
 'NDL_8': ('leaf', 'IS', 'LOCATION_STATUS', None),
 'NDL_9': ('leaf', 'IS', 'PATIENT_LOCATION_TYPE', None),
 'NR': ('sequence', (('NR_1', (0, 1)), ('NR_2', (0, 1)))),
 'NR_1': ('leaf', 'NM', 'LOW_VALUE', None),
 'NR_2': ('leaf', 'NM', 'HIGH_VALUE', None),
 'OCD': ('sequence', (('OCD_1', (1, 1)), ('OCD_2', (1, 1)))),
 'OCD_1': ('leaf', 'CNE', 'OCCURRENCE_CODE', None),
 'OCD_2': ('leaf', 'DT', 'OCCURRENCE_DATE', None),
 'OSD': ('sequence',
         (('OSD_1', (1, 1)),
          ('OSD_2', (1, 1)),
          ('OSD_3', (0, 1)),
          ('OSD_4', (1, 1)),
          ('OSD_5', (0, 1)),
          ('OSD_6', (0, 1)),
          ('OSD_7', (0, 1)),
          ('OSD_8', (1, 1)),
          ('OSD_9', (0, 1)),
          ('OSD_10', (1, 1)),
          ('OSD_11', (0, 1)))),
 'OSD_1': ('leaf', 'ID', 'SEQUENCE_RESULTS_FLAG', None),
 'OSD_10': ('leaf', 'ST', 'FILLER_ORDER_NUMBER_UNIVERSAL_ID', None),
 'OSD_11': ('leaf', 'ID', 'FILLER_ORDER_NUMBER_UNIVERSAL_ID_TYPE', None),
 'OSD_2': ('leaf', 'ST', 'PLACER_ORDER_NUMBER_ENTITY_IDENTIFIER', None),
 'OSD_3': ('leaf', 'IS', 'PLACER_ORDER_NUMBER_NAMESPACE_ID', None),
 'OSD_4': ('leaf', 'ST', 'FILLER_ORDER_NUMBER_ENTITY_IDENTIFIER', None),
 'OSD_5': ('leaf', 'IS', 'FILLER_ORDER_NUMBER_NAMESPACE_ID', None),
 'OSD_6': ('leaf', 'ST', 'SEQUENCE_CONDITION_VALUE', None),
 'OSD_7': ('leaf', 'NM', 'MAXIMUM_NUMBER_OF_REPEATS', None),
 'OSD_8': ('leaf', 'ST', 'PLACER_ORDER_NUMBER_UNIVERSAL_ID', None),
 'OSD_9': ('leaf', 'ID', 'PLACER_ORDER_NUMBER_UNIVERSAL_ID_TYPE', None),
 'OSP': ('sequence',
         (('OSP_1', (1, 1)), ('OSP_2', (0, 1)), ('OSP_3', (0, 1)))),
 'OSP_1': ('leaf', 'CNE', 'OCCURRENCE_SPAN_CODE', None),
 'OSP_2': ('leaf', 'DT', 'OCCURRENCE_SPAN_START_DATE', None),
 'OSP_3': ('leaf', 'DT', 'OCCURRENCE_SPAN_STOP_DATE', None),
 'PIP': ('sequence',
         (('PIP_1', (1, 1)),
          ('PIP_2', (0, 1)),
          ('PIP_3', (0, 1)),
          ('PIP_4', (0, 1)),
          ('PIP_5', (0, 1)))),
 'PIP_1': ('leaf', 'CWE', 'PRIVILEGE', None),
 'PIP_2': ('leaf', 'CWE', 'PRIVILEGE_CLASS', None),
 'PIP_3': ('leaf', 'DT', 'EXPIRATION_DATE', None),
 'PIP_4': ('leaf', 'DT', 'ACTIVATION_DATE', None),
 'PIP_5': ('leaf', 'EI', 'FACILITY', None),
 'PL': ('sequence',
        (('PL_1', (0, 1)),
         ('PL_2', (0, 1)),
         ('PL_3', (0, 1)),
         ('PL_4', (0, 1)),
         ('PL_5', (0, 1)),
         ('PL_6', (0, 1)),
         ('PL_7', (0, 1)),
         ('PL_8', (0, 1)),
         ('PL_9', (0, 1)),
         ('PL_10', (0, 1)),
         ('PL_11', (0, 1)))),
 'PLN': ('sequence',
         (('PLN_1', (1, 1)),
          ('PLN_2', (1, 1)),
          ('PLN_3', (0, 1)),
          ('PLN_4', (0, 1)))),
 'PLN_1': ('leaf', 'ST', 'ID_NUMBER', None),
 'PLN_2': ('leaf', 'IS', 'TYPE_OF_ID_NUMBER', None),
 'PLN_3': ('leaf', 'ST', 'STATE_OTHER_QUALIFYING_INFORMATION', None),
 'PLN_4': ('leaf', 'DT', 'EXPIRATION_DATE', None),
 'PL_1': ('leaf', 'IS', 'POINT_OF_CARE', None),
 'PL_10': ('leaf', 'EI', 'COMPREHENSIVE_LOCATION_IDENTIFIER', None),
 'PL_11': ('leaf', 'HD', 'ASSIGNING_AUTHORITY_FOR_LOCATION', None),
 'PL_2': ('leaf', 'IS', 'ROOM', None),
 'PL_3': ('leaf', 'IS', 'BED', None),
 'PL_4': ('leaf', 'HD', 'FACILITY', None),
 'PL_5': ('leaf', 'IS', 'LOCATION_STATUS', None),
 'PL_6': ('leaf', 'IS', 'PERSON_LOCATION_TYPE', None),
 'PL_7': ('leaf', 'IS', 'BUILDING', None),
 'PL_8': ('leaf', 'IS', 'FLOOR', None),
 'PL_9': ('leaf', 'ST', 'LOCATION_DESCRIPTION', None),
 'PPN': ('sequence',
         (('PPN_1', (0, 1)),
          ('PPN_2', (0, 1)),
          ('PPN_3', (0, 1)),
          ('PPN_4', (0, 1)),
          ('PPN_5', (0, 1)),
          ('PPN_6', (0, 1)),
          ('PPN_7', (0, 1)),
          ('PPN_8', (0, 1)),
          ('PPN_9', (0, 1)),
          ('PPN_10', (0, 1)),
          ('PPN_11', (0, 1)),
          ('PPN_12', (0, 1)),
          ('PPN_13', (0, 1)),
          ('PPN_14', (0, 1)),
          ('PPN_15', (0, 1)),
          ('PPN_16', (0, 1)),
          ('PPN_17', (0, 1)),
          ('PPN_18', (0, 1)),
          ('PPN_19', (0, 1)),
          ('PPN_20', (0, 1)),
          ('PPN_21', (0, 1)),
          ('PPN_22', (0, 1)),
          ('PPN_23', (0, 1)),
          ('PPN_24', (0, 1)))),
 'PPN_1': ('leaf', 'ST', 'ID_NUMBER', None),
 'PPN_10': ('leaf', 'ID', 'NAME_TYPE_CODE', None),
 'PPN_11': ('leaf', 'ST', 'IDENTIFIER_CHECK_DIGIT', None),
 'PPN_12': ('leaf', 'ID', 'CHECK_DIGIT_SCHEME', None),
 'PPN_13': ('leaf', 'ID', 'IDENTIFIER_TYPE_CODE', None),
 'PPN_14': ('leaf', 'HD', 'ASSIGNING_FACILITY', None),
 'PPN_15': ('leaf', 'DTM', 'DATE_TIME_ACTION_PERFORMED', None),
 'PPN_16': ('leaf', 'ID', 'NAME_REPRESENTATION_CODE', None),
 'PPN_17': ('leaf', 'CWE', 'NAME_CONTEXT', None),
 'PPN_18': ('leaf', 'DR', 'NAME_VALIDITY_RANGE', None),
 'PPN_19': ('leaf', 'ID', 'NAME_ASSEMBLY_ORDER', None),
 'PPN_2': ('leaf', 'FN', 'FAMILY_NAME', None),
 'PPN_20': ('leaf', 'DTM', 'EFFECTIVE_DATE', None),
 'PPN_21': ('leaf', 'DTM', 'EXPIRATION_DATE', None),
 'PPN_22': ('leaf', 'ST', 'PROFESSIONAL_SUFFIX', None),
 'PPN_23': ('leaf', 'CWE', 'ASSIGNING_JURISDICTION', None),
 'PPN_24': ('leaf', 'CWE', 'ASSIGNING_AGENCY_OR_DEPARTMENT', None),
 'PPN_3': ('leaf', 'ST', 'GIVEN_NAME', None),
 'PPN_4': ('leaf',
           'ST',
           'SECOND_AND_FURTHER_GIVEN_NAMES_OR_INITIALS_THEREOF',
           None),
 'PPN_5': ('leaf', 'ST', 'SUFFIX_E_G_JR_OR_III_', None),
 'PPN_6': ('leaf', 'ST', 'PREFIX_E_G_DR_', None),
 'PPN_7': ('leaf', 'IS', 'DEGREE_E_G_MD_', None),
 'PPN_8': ('leaf', 'IS', 'SOURCE_TABLE', None),
 'PPN_9': ('leaf', 'HD', 'ASSIGNING_AUTHORITY', None),
 'PRL': ('sequence',
         (('PRL_1', (1, 1)), ('PRL_2', (0, 1)), ('PRL_3', (0, 1)))),
 'PRL_1': ('leaf', 'CWE', 'PARENT_OBSERVATION_IDENTIFIER', None),
 'PRL_2': ('leaf', 'ST', 'PARENT_OBSERVATION_SUB_IDENTIFIER', None),
 'PRL_3': ('leaf', 'TX', 'PARENT_OBSERVATION_VALUE_DESCRIPTOR', None),
 'PT': ('sequence', (('PT_1', (0, 1)), ('PT_2', (0, 1)))),
 'PTA': ('sequence',
         (('PTA_1', (1, 1)),
          ('PTA_2', (0, 1)),
          ('PTA_3', (0, 1)),
          ('PTA_4', (1, 1)))),
 'PTA_1': ('leaf', 'IS', 'POLICY_TYPE', None),
 'PTA_2': ('leaf', 'IS', 'AMOUNT_CLASS', None),
 'PTA_3': ('leaf', 'NM', 'MONEY_OR_PERCENTAGE_QUANTITY', None),
 'PTA_4': ('leaf', 'MOP', 'MONEY_OR_PERCENTAGE', None),
 'PT_1': ('leaf', 'ID', 'PROCESSING_ID', None),
 'PT_2': ('leaf', 'ID', 'PROCESSING_MODE', None),
 'QIP': ('sequence', (('QIP_1', (1, 1)), ('QIP_2', (1, 1)))),
 'QIP_1': ('leaf', 'ST', 'SEGMENT_FIELD_NAME', None),
 'QIP_2': ('leaf', 'ST', 'VALUES', None),
 'QSC': ('sequence',
         (('QSC_1', (1, 1)),
          ('QSC_2', (0, 1)),
          ('QSC_3', (0, 1)),
          ('QSC_4', (0, 1)))),
 'QSC_1': ('leaf', 'ST', 'SEGMENT_FIELD_NAME', None),
 'QSC_2': ('leaf', 'ID', 'RELATIONAL_OPERATOR', None),
 'QSC_3': ('leaf', 'ST', 'VALUE', None),
 'QSC_4': ('leaf', 'ID', 'RELATIONAL_CONJUNCTION', None),
 'RCD': ('sequence',
         (('RCD_1', (0, 1)), ('RCD_2', (0, 1)), ('RCD_3', (0, 1)))),
 'RCD_1': ('leaf', 'ST', 'SEGMENT_FIELD_NAME', None),
 'RCD_2': ('leaf', 'ID', 'HL7_DATA_TYPE', None),
 'RCD_3': ('leaf', 'NM', 'MAXIMUM_COLUMN_WIDTH', None),
 'RFR': ('sequence',
         (('RFR_1', (1, 1)),
          ('RFR_2', (0, 1)),
          ('RFR_3', (0, 1)),
          ('RFR_4', (0, 1)),
          ('RFR_5', (0, 1)),
          ('RFR_6', (0, 1)),
          ('RFR_7', (0, 1)))),
 'RFR_1': ('leaf', 'NR', 'NUMERIC_RANGE', None),
 'RFR_2': ('leaf', 'IS', 'ADMINISTRATIVE_SEX', None),
 'RFR_3': ('leaf', 'NR', 'AGE_RANGE', None),
 'RFR_4': ('leaf', 'NR', 'GESTATIONAL_AGE_RANGE', None),
 'RFR_5': ('leaf', 'ST', 'SPECIES', None),
 'RFR_6': ('leaf', 'ST', 'RACE_SUBSPECIES', None),
 'RFR_7': ('leaf', 'TX', 'CONDITIONS', None),
 'RI': ('sequence', (('RI_1', (0, 1)), ('RI_2', (0, 1)))),
 'RI_1': ('leaf', 'IS', 'REPEAT_PATTERN', None),
 'RI_2': ('leaf', 'ST', 'EXPLICIT_TIME_INTERVAL', None),
 'RMC': ('sequence',
         (('RMC_1', (1, 1)),
          ('RMC_2', (0, 1)),
          ('RMC_3', (0, 1)),
          ('RMC_4', (1, 1)))),
 'RMC_1': ('leaf', 'IS', 'ROOM_TYPE', None),
 'RMC_2': ('leaf', 'IS', 'AMOUNT_TYPE', None),
 'RMC_3': ('leaf', 'NM', 'COVERAGE_AMOUNT', None),
 'RMC_4': ('leaf', 'MOP', 'MONEY_OR_PERCENTAGE', None),
 'RP': ('sequence',
        (('RP_1', (0, 1)),
         ('RP_2', (0, 1)),
         ('RP_3', (0, 1)),
         ('RP_4', (0, 1)))),
 'RPT': ('sequence',
         (('RPT_1', (1, 1)),
          ('RPT_2', (0, 1)),
          ('RPT_3', (0, 1)),
          ('RPT_4', (0, 1)),
          ('RPT_5', (0, 1)),
          ('RPT_6', (0, 1)),
          ('RPT_7', (0, 1)),
          ('RPT_8', (0, 1)),
          ('RPT_9', (0, 1)),
          ('RPT_10', (0, 1)),
          ('RPT_11', (0, 1)))),
 'RPT_1': ('leaf', 'CWE', 'REPEAT_PATTERN_CODE', None),
 'RPT_10': ('leaf', 'IS', 'EVENT_OFFSET_UNITS', None),
 'RPT_11': ('leaf', 'GTS', 'GENERAL_TIMING_SPECIFICATION', None),
 'RPT_2': ('leaf', 'ID', 'CALENDAR_ALIGNMENT', None),
 'RPT_3': ('leaf', 'NM', 'PHASE_RANGE_BEGIN_VALUE', None),
 'RPT_4': ('leaf', 'NM', 'PHASE_RANGE_END_VALUE', None),
 'RPT_5': ('leaf', 'NM', 'PERIOD_QUANTITY', None),
 'RPT_6': ('leaf', 'IS', 'PERIOD_UNITS', None),
 'RPT_7': ('leaf', 'ID', 'INSTITUTION_SPECIFIED_TIME', None),
 'RPT_8': ('leaf', 'ID', 'EVENT', None),
 'RPT_9': ('leaf', 'NM', 'EVENT_OFFSET_QUANTITY', None),
 'RP_1': ('leaf', 'ST', 'POINTER', None),
 'RP_2': ('leaf', 'HD', 'APPLICATION_ID', None),
 'RP_3': ('leaf', 'ID', 'TYPE_OF_DATA', None),
 'RP_4': ('leaf', 'ID', 'SUBTYPE', None),
 'SAD': ('sequence',
         (('SAD_1', (0, 1)), ('SAD_2', (0, 1)), ('SAD_3', (0, 1)))),
 'SAD_1': ('leaf', 'ST', 'STREET_OR_MAILING_ADDRESS', None),
 'SAD_2': ('leaf', 'ST', 'STREET_NAME', None),
 'SAD_3': ('leaf', 'ST', 'DWELLING_NUMBER', None),
 'SCV': ('sequence', (('SCV_1', (0, 1)), ('SCV_2', (0, 1)))),
 'SCV_1': ('leaf', 'CWE', 'PARAMETER_CLASS', None),
 'SCV_2': ('leaf', 'ST', 'PARAMETER_VALUE', None),
 'SN': ('sequence',
        (('SN_1', (0, 1)),
         ('SN_2', (0, 1)),
         ('SN_3', (0, 1)),
         ('SN_4', (0, 1)))),
 'SN_1': ('leaf', 'ST', 'COMPARATOR', None),
 'SN_2': ('leaf', 'NM', 'NUM1', None),
 'SN_3': ('leaf', 'ST', 'SEPARATOR_SUFFIX', None),
 'SN_4': ('leaf', 'NM', 'NUM2', None),
 'SPD': ('sequence',
         (('SPD_1', (1, 1)),
          ('SPD_2', (0, 1)),
          ('SPD_3', (0, 1)),
          ('SPD_4', (0, 1)))),
 'SPD_1': ('leaf', 'ST', 'SPECIALTY_NAME', None),
 'SPD_2': ('leaf', 'ST', 'GOVERNING_BOARD', None),
 'SPD_3': ('leaf', 'ID', 'ELIGIBLE_OR_CERTIFIED', None),
 'SPD_4': ('leaf', 'DT', 'DATE_OF_CERTIFICATION', None),
 'SPS': ('sequence',
         (('SPS_1', (0, 1)),
          ('SPS_2', (0, 1)),
          ('SPS_3', (0, 1)),
          ('SPS_4', (0, 1)),
          ('SPS_5', (0, 1)),
          ('SPS_6', (0, 1)),
          ('SPS_7', (0, 1)))),
 'SPS_1': ('leaf', 'CWE', 'SPECIMEN_SOURCE_NAME_OR_CODE', None),
 'SPS_2': ('leaf', 'CWE', 'ADDITIVES', None),
 'SPS_3': ('leaf', 'TX', 'SPECIMEN_COLLECTION_METHOD', None),
 'SPS_4': ('leaf', 'CWE', 'BODY_SITE', None),
 'SPS_5': ('leaf', 'CWE', 'SITE_MODIFIER', None),
 'SPS_6': ('leaf', 'CWE', 'COLLECTION_METHOD_MODIFIER_CODE', None),
 'SPS_7': ('leaf', 'CWE', 'SPECIMEN_ROLE', None),
 'SRT': ('sequence', (('SRT_1', (1, 1)), ('SRT_2', (0, 1)))),
 'SRT_1': ('leaf', 'ST', 'SORT_BY_FIELD', None),
 'SRT_2': ('leaf', 'ID', 'SEQUENCING', None),
 'TQ': ('sequence',
        (('TQ_1', (0, 1)),
         ('TQ_2', (0, 1)),
         ('TQ_3', (0, 1)),
         ('TQ_4', (0, 1)),
         ('TQ_5', (0, 1)),
         ('TQ_6', (0, 1)),
         ('TQ_7', (0, 1)),
         ('TQ_8', (0, 1)),
         ('TQ_9', (0, 1)),
         ('TQ_10', (0, 1)),
         ('TQ_11', (0, 1)),
         ('TQ_12', (0, 1)))),
 'TQ_1': ('leaf', 'CQ', 'QUANTITY', None),
 'TQ_10': ('leaf', 'OSD', 'ORDER_SEQUENCING', None),
 'TQ_11': ('leaf', 'CWE', 'OCCURRENCE_DURATION', None),
 'TQ_12': ('leaf', 'NM', 'TOTAL_OCCURRENCES', None),
 'TQ_2': ('leaf', 'RI', 'INTERVAL', None),
 'TQ_3': ('leaf', 'ST', 'DURATION', None),
 'TQ_4': ('leaf', 'DTM', 'START_DATE_TIME', None),
 'TQ_5': ('leaf', 'DTM', 'END_DATE_TIME', None),
 'TQ_6': ('leaf', 'ST', 'PRIORITY', None),
 'TQ_7': ('leaf', 'ST', 'CONDITION', None),
 'TQ_8': ('leaf', 'TX', 'TEXT', None),
 'TQ_9': ('leaf', 'ID', 'CONJUNCTION', None),
 'UVC': ('sequence', (('UVC_1', (1, 1)), ('UVC_2', (0, 1)))),
 'UVC_1': ('leaf', 'CNE', 'VALUE_CODE', None),
 'UVC_2': ('leaf', 'MO', 'VALUE_AMOUNT', None),
 'VARIES': None,
 'VH': ('sequence',
        (('VH_1', (0, 1)),
         ('VH_2', (0, 1)),
         ('VH_3', (0, 1)),
         ('VH_4', (0, 1)))),
 'VH_1': ('leaf', 'ID', 'START_DAY_RANGE', None),
 'VH_2': ('leaf', 'ID', 'END_DAY_RANGE', None),
 'VH_3': ('leaf', 'TM', 'START_HOUR_RANGE', None),
 'VH_4': ('leaf', 'TM', 'END_HOUR_RANGE', None),
 'VID': ('sequence',
         (('VID_1', (0, 1)), ('VID_2', (0, 1)), ('VID_3', (0, 1)))),
 'VID_1': ('leaf', 'ID', 'VERSION_ID', None),
 'VID_2': ('leaf', 'CWE', 'INTERNATIONALIZATION_CODE', None),
 'VID_3': ('leaf', 'CWE', 'INTERNATIONAL_VERSION_ID', None),
 'VR': ('sequence', (('VR_1', (0, 1)), ('VR_2', (0, 1)))),
 'VR_1': ('leaf', 'ST', 'FIRST_DATA_CODE_VALUE', None),
 'VR_2': ('leaf', 'ST', 'LAST_DATA_CODE_VALUE', None),
 'WVI': ('sequence', (('WVI_1', (1, 1)), ('WVI_2', (0, 1)))),
 'WVI_1': ('leaf', 'NM', 'CHANNEL_NUMBER', None),
 'WVI_2': ('leaf', 'ST', 'CHANNEL_NAME', None),
 'WVS': ('sequence', (('WVS_1', (1, 1)), ('WVS_2', (0, 1)))),
 'WVS_1': ('leaf', 'ST', 'SOURCE_ONE_NAME', None),
 'WVS_2': ('leaf', 'ST', 'SOURCE_TWO_NAME', None),
 'XAD': ('sequence',
         (('XAD_1', (0, 1)),
          ('XAD_2', (0, 1)),
          ('XAD_3', (0, 1)),
          ('XAD_4', (0, 1)),
          ('XAD_5', (0, 1)),
          ('XAD_6', (0, 1)),
          ('XAD_7', (0, 1)),
          ('XAD_8', (0, 1)),
          ('XAD_9', (0, 1)),
          ('XAD_10', (0, 1)),
          ('XAD_11', (0, 1)),
          ('XAD_12', (0, 1)),
          ('XAD_13', (0, 1)),
          ('XAD_14', (0, 1)),
          ('XAD_15', (0, 1)),
          ('XAD_16', (0, 1)),
          ('XAD_17', (0, 1)),
          ('XAD_18', (0, 1)),
          ('XAD_19', (0, 1)),
          ('XAD_20', (0, 1)),
          ('XAD_21', (0, 1)),
          ('XAD_22', (0, 1)),
          ('XAD_23', (0, 1)))),
 'XAD_1': ('leaf', 'SAD', 'STREET_ADDRESS', None),
 'XAD_10': ('leaf', 'IS', 'CENSUS_TRACT', None),
 'XAD_11': ('leaf', 'ID', 'ADDRESS_REPRESENTATION_CODE', None),
 'XAD_12': ('leaf', 'DR', 'ADDRESS_VALIDITY_RANGE', None),
 'XAD_13': ('leaf', 'DTM', 'EFFECTIVE_DATE', None),
 'XAD_14': ('leaf', 'DTM', 'EXPIRATION_DATE', None),
 'XAD_15': ('leaf', 'CWE', 'EXPIRATION_REASON', None),
 'XAD_16': ('leaf', 'ID', 'TEMPORARY_INDICATOR', None),
 'XAD_17': ('leaf', 'ID', 'BAD_ADDRESS_INDICATOR', None),
 'XAD_18': ('leaf', 'ID', 'ADDRESS_USAGE', None),
 'XAD_19': ('leaf', 'ST', 'ADDRESSEE', None),
 'XAD_2': ('leaf', 'ST', 'OTHER_DESIGNATION', None),
 'XAD_20': ('leaf', 'ST', 'COMMENT', None),
 'XAD_21': ('leaf', 'NM', 'PREFERENCE_ORDER', None),
 'XAD_22': ('leaf', 'CWE', 'PROTECTION_CODE', None),
 'XAD_23': ('leaf', 'EI', 'ADDRESS_IDENTIFIER', None),
 'XAD_3': ('leaf', 'ST', 'CITY', None),
 'XAD_4': ('leaf', 'ST', 'STATE_OR_PROVINCE', None),
 'XAD_5': ('leaf', 'ST', 'ZIP_OR_POSTAL_CODE', None),
 'XAD_6': ('leaf', 'ID', 'COUNTRY', None),
 'XAD_7': ('leaf', 'ID', 'ADDRESS_TYPE', None),
 'XAD_8': ('leaf', 'ST', 'OTHER_GEOGRAPHIC_DESIGNATION', None),
 'XAD_9': ('leaf', 'IS', 'COUNTY_PARISH_CODE', None),
 'XCN': ('sequence',
         (('XCN_1', (0, 1)),
          ('XCN_2', (0, 1)),
          ('XCN_3', (0, 1)),
          ('XCN_4', (0, 1)),
          ('XCN_5', (0, 1)),
          ('XCN_6', (0, 1)),
          ('XCN_7', (0, 1)),
          ('XCN_8', (0, 1)),
          ('XCN_9', (0, 1)),
          ('XCN_10', (0, 1)),
          ('XCN_11', (0, 1)),
          ('XCN_12', (0, 1)),
          ('XCN_13', (0, 1)),
          ('XCN_14', (0, 1)),
          ('XCN_15', (0, 1)),
          ('XCN_16', (0, 1)),
          ('XCN_17', (0, 1)),
          ('XCN_18', (0, 1)),
          ('XCN_19', (0, 1)),
          ('XCN_20', (0, 1)),
          ('XCN_21', (0, 1)),
          ('XCN_22', (0, 1)),
          ('XCN_23', (0, 1)))),
 'XCN_1': ('leaf', 'ST', 'ID_NUMBER', None),
 'XCN_10': ('leaf', 'ID', 'NAME_TYPE_CODE', None),
 'XCN_11': ('leaf', 'ST', 'IDENTIFIER_CHECK_DIGIT', None),
 'XCN_12': ('leaf', 'ID', 'CHECK_DIGIT_SCHEME', None),
 'XCN_13': ('leaf', 'ID', 'IDENTIFIER_TYPE_CODE', None),
 'XCN_14': ('leaf', 'HD', 'ASSIGNING_FACILITY', None),
 'XCN_15': ('leaf', 'ID', 'NAME_REPRESENTATION_CODE', None),
 'XCN_16': ('leaf', 'CWE', 'NAME_CONTEXT', None),
 'XCN_17': ('leaf', 'DR', 'NAME_VALIDITY_RANGE', None),
 'XCN_18': ('leaf', 'ID', 'NAME_ASSEMBLY_ORDER', None),
 'XCN_19': ('leaf', 'DTM', 'EFFECTIVE_DATE', None),
 'XCN_2': ('leaf', 'FN', 'FAMILY_NAME', None),
 'XCN_20': ('leaf', 'DTM', 'EXPIRATION_DATE', None),
 'XCN_21': ('leaf', 'ST', 'PROFESSIONAL_SUFFIX', None),
 'XCN_22': ('leaf', 'CWE', 'ASSIGNING_JURISDICTION', None),
 'XCN_23': ('leaf', 'CWE', 'ASSIGNING_AGENCY_OR_DEPARTMENT', None),
 'XCN_3': ('leaf', 'ST', 'GIVEN_NAME', None),
 'XCN_4': ('leaf',
           'ST',
           'SECOND_AND_FURTHER_GIVEN_NAMES_OR_INITIALS_THEREOF',
           None),
 'XCN_5': ('leaf', 'ST', 'SUFFIX_E_G_JR_OR_III_', None),
 'XCN_6': ('leaf', 'ST', 'PREFIX_E_G_DR_', None),
 'XCN_7': ('leaf', 'IS', 'DEGREE_E_G_MD_', None),
 'XCN_8': ('leaf', 'IS', 'SOURCE_TABLE', None),
 'XCN_9': ('leaf', 'HD', 'ASSIGNING_AUTHORITY', None),
 'XON': ('sequence',
         (('XON_1', (0, 1)),
          ('XON_2', (0, 1)),
          ('XON_3', (0, 1)),
          ('XON_4', (0, 1)),
          ('XON_5', (0, 1)),
          ('XON_6', (0, 1)),
          ('XON_7', (0, 1)),
          ('XON_8', (0, 1)),
          ('XON_9', (0, 1)),
          ('XON_10', (0, 1)))),
 'XON_1': ('leaf', 'ST', 'ORGANIZATION_NAME', None),
 'XON_10': ('leaf', 'ST', 'ORGANIZATION_IDENTIFIER', None),
 'XON_2': ('leaf', 'IS', 'ORGANIZATION_NAME_TYPE_CODE', None),
 'XON_3': ('leaf', 'NM', 'ID_NUMBER', None),
 'XON_4': ('leaf', 'NM', 'IDENTIFIER_CHECK_DIGIT', None),
 'XON_5': ('leaf', 'ID', 'CHECK_DIGIT_SCHEME', None),
 'XON_6': ('leaf', 'HD', 'ASSIGNING_AUTHORITY', None),
 'XON_7': ('leaf', 'ID', 'IDENTIFIER_TYPE_CODE', None),
 'XON_8': ('leaf', 'HD', 'ASSIGNING_FACILITY', None),
 'XON_9': ('leaf', 'ID', 'NAME_REPRESENTATION_CODE', None),
 'XPN': ('sequence',
         (('XPN_1', (0, 1)),
          ('XPN_2', (0, 1)),
          ('XPN_3', (0, 1)),
          ('XPN_4', (0, 1)),
          ('XPN_5', (0, 1)),
          ('XPN_6', (0, 1)),
          ('XPN_7', (0, 1)),
          ('XPN_8', (0, 1)),
          ('XPN_9', (0, 1)),
          ('XPN_10', (0, 1)),
          ('XPN_11', (0, 1)),
          ('XPN_12', (0, 1)),
          ('XPN_13', (0, 1)),
          ('XPN_14', (0, 1)))),
 'XPN_1': ('leaf', 'FN', 'FAMILY_NAME', None),
 'XPN_10': ('leaf', 'DR', 'NAME_VALIDITY_RANGE', None),
 'XPN_11': ('leaf', 'ID', 'NAME_ASSEMBLY_ORDER', None),
 'XPN_12': ('leaf', 'DTM', 'EFFECTIVE_DATE', None),
 'XPN_13': ('leaf', 'DTM', 'EXPIRATION_DATE', None),
 'XPN_14': ('leaf', 'ST', 'PROFESSIONAL_SUFFIX', None),
 'XPN_2': ('leaf', 'ST', 'GIVEN_NAME', None),
 'XPN_3': ('leaf',
           'ST',
           'SECOND_AND_FURTHER_GIVEN_NAMES_OR_INITIALS_THEREOF',
           None),
 'XPN_4': ('leaf', 'ST', 'SUFFIX_E_G_JR_OR_III_', None),
 'XPN_5': ('leaf', 'ST', 'PREFIX_E_G_DR_', None),
 'XPN_6': ('leaf', 'IS', 'DEGREE_E_G_MD_', None),
 'XPN_7': ('leaf', 'ID', 'NAME_TYPE_CODE', None),
 'XPN_8': ('leaf', 'ID', 'NAME_REPRESENTATION_CODE', None),
 'XPN_9': ('leaf', 'CWE', 'NAME_CONTEXT', None),
 'XTN': ('sequence',
         (('XTN_2', (0, 1)),
          ('XTN_3', (0, 1)),
          ('XTN_4', (0, 1)),
          ('XTN_5', (0, 1)),
          ('XTN_6', (0, 1)),
          ('XTN_7', (0, 1)),
          ('XTN_8', (0, 1)),
          ('XTN_9', (0, 1)),
          ('XTN_10', (0, 1)),
          ('XTN_11', (0, 1)),
          ('XTN_12', (0, 1)),
          ('XTN_13', (0, 1)),
          ('XTN_14', (0, 1)),
          ('XTN_15', (0, 1)),
          ('XTN_16', (0, 1)),
          ('XTN_17', (0, 1)),
          ('XTN_18', (0, 1)))),
 'XTN_10': ('leaf', 'ST', 'EXTENSION_PREFIX', None),
 'XTN_11': ('leaf', 'ST', 'SPEED_DIAL_CODE', None),
 'XTN_12': ('leaf', 'ST', 'UNFORMATTED_TELEPHONE_NUMBER', None),
 'XTN_13': ('leaf', 'DTM', 'EFFECTIVE_START_DATE', None),
 'XTN_14': ('leaf', 'DTM', 'EXPIRATION_DATE', None),
 'XTN_15': ('leaf', 'CWE', 'EXPIRATION_REASON', None),
 'XTN_16': ('leaf', 'CWE', 'PROTECTION_CODE', None),
 'XTN_17': ('leaf', 'EI', 'SHARED_TELECOMMUNICATION_IDENTIFIER', None),
 'XTN_18': ('leaf', 'NM', 'PREFERENCE_ORDER', None),
 'XTN_2': ('leaf', 'ID', 'TELECOMMUNICATION_USE_CODE', None),
 'XTN_3': ('leaf', 'ID', 'TELECOMMUNICATION_EQUIPMENT_TYPE', None),
 'XTN_4': ('leaf', 'ST', 'COMMUNICATION_ADDRESS', None),
 'XTN_5': ('leaf', 'NM', 'COUNTRY_CODE', None),
 'XTN_6': ('leaf', 'NM', 'AREA_CITY_CODE', None),
 'XTN_7': ('leaf', 'NM', 'LOCAL_NUMBER', None),
 'XTN_8': ('leaf', 'NM', 'EXTENSION', None),
 'XTN_9': ('leaf', 'ST', 'ANY_TEXT', None)}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
HL7apy - validation
"""

from hl7apy import load_reference
from hl7apy.consts import VALIDATION_LEVEL

class Validator(object):
    """
    Class that handles validation. It defines validation levels and validate
    an element using :attr:`hl7apy.consts.VALIDATION_LEVEL.STRICT` validation level
    """
    def __init__(self, level):
        super(Validator, self).__init__()
        self.level = level

    @staticmethod
    def validate(element):
        """
        Checks if the :class:`hl7apy.core.Element` is a valid HL7 message according to the official structures.
        It checks if it follows the max and min number of occurrences for every child and if it
        doesn't contain children not allowed.

        :param element: :class:`hl7apy.core.Element`
        :return: The element to be validated
        """

        from hl7apy.core import is_base_datatype

        def _is_valid(ref, element):
            children = set([ child.name for child in element.children ])
            valid_children = set([ child[0] for child in ref[1] ])
            if not children <= valid_children:
                return False
            for child in ref[1]:
                child_name, cardinality = child
                min, max = cardinality

                children = element.children.get(child_name)
                num_children = len(children)
                if max != -1:
                    if num_children < min or num_children > max:
                        return False
                else:
                    if num_children < min:
                        return False
                for el in children:
                    if not Validator.validate(el):
                        return False
            return True

        if element.is_unknown():
            return False
        ref = load_reference(element.name, element.classname, element.version)
        if ref[0] in ('sequence', 'choice'):
            return _is_valid(ref, element)
        else: # it's an element that has a datatype
            if element.datatype != ref[1]:
                return False
            if element.is_unknown():
                return False
            if not is_base_datatype(element.datatype, element.version):
                # Component just to search in the datatypes....
                ref = load_reference(element.datatype, 'Component', element.version)
                return _is_valid(ref, element)

        return True

    @staticmethod
    def is_strict(level):
        """
        Check if the given validation level is strict

        :type level: ``int``
        :param level: validation level (see `hl7apy.consts.VALIDATION_LEVEL`)
        :rtype: ``bool``
        :return: ``True`` if validation level is strict
        """
        return level == VALIDATION_LEVEL.STRICT

    @staticmethod
    def is_quiet(level):
        """
        Check if the given validation level is quiet

        :type level: ``int``
        :param level: validation level (see `hl7apy.consts.VALIDATION_LEVEL`)
        :rtype: ``bool``
        :return: ``True`` if validation level is quiet
        """
        return level == VALIDATION_LEVEL.QUIET
